# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import logging
import secrets
import warnings
from typing import TYPE_CHECKING, Any

from onetl._util.spark import override_job_description
from onetl._util.sql import clear_statement
from onetl.connection.db_connection.db_connection import DBConnection
from onetl.connection.db_connection.jdbc_connection.dialect import JDBCDialect
from onetl.connection.db_connection.jdbc_connection.options import (
    JDBCLegacyOptions,
    JDBCPartitioningMode,
    JDBCReadOptions,
    JDBCSQLOptions,
    JDBCTableExistBehavior,
    JDBCWriteOptions,
)
from onetl.connection.db_connection.jdbc_mixin import JDBCMixin
from onetl.connection.db_connection.jdbc_mixin.options import JDBCFetchOptions
from onetl.hooks import slot, support_hooks
from onetl.hwm import Window
from onetl.log import log_lines, log_with_indent

if TYPE_CHECKING:
    from pyspark.sql import DataFrame
    from pyspark.sql.types import StructType

log = logging.getLogger(__name__)

# parameters accepted by spark.read.jdbc method:
#  spark.read.jdbc(
#    url, table, column, lowerBound, upperBound, numPartitions, predicates
#    properties:  { "user" : "SYSTEM", "password" : "mypassword", ... })
READ_TOP_LEVEL_OPTIONS = frozenset(("url", "column", "lower_bound", "upper_bound", "num_partitions", "predicates"))

# parameters accepted by spark.write.jdbc method:
#   spark.write.jdbc(
#     url, table, mode,
#     properties:  { "user" : "SYSTEM", "password" : "mypassword", ... })
WRITE_TOP_LEVEL_OPTIONS = frozenset("url")


@support_hooks
class JDBCConnection(JDBCMixin, DBConnection):
    Dialect = JDBCDialect
    ReadOptions = JDBCReadOptions
    SQLOptions = JDBCSQLOptions
    WriteOptions = JDBCWriteOptions
    Options = JDBCLegacyOptions

    @slot
    def sql(
        self,
        query: str,
        options: JDBCSQLOptions | dict | None = None,
    ) -> DataFrame:
        """
        **Lazily** execute SELECT statement **on Spark executor** and return DataFrame. |support_hooks|

        Same as ``spark.read.jdbc(query)``.

        .. versionadded:: 0.2.0

        Parameters
        ----------
        query : str

            SQL query to be executed.

        options : dict, :obj:`~SQLOptions`, default: ``None``

            Spark options to be used while fetching data.

        Returns
        -------
        df : pyspark.sql.dataframe.DataFrame

            Spark dataframe

        """

        if isinstance(options, JDBCReadOptions):
            msg = "Using `ReadOptions` for `sql` method is deprecated, use `SQLOptions` instead."
            warnings.warn(msg, UserWarning, stacklevel=3)
            options = self.SQLOptions.parse_obj(options.dict(exclude={"partitioning_mode"}, exclude_none=True))

        query = clear_statement(query)

        log.info("|%s| Detected dialect: '%s'", self.__class__.__name__, self._get_spark_dialect_name())
        log.info("|%s| Executing SQL query (on executor):", self.__class__.__name__)
        log_lines(log, query)

        try:
            with override_job_description(self.spark, f"{self}.sql()"):
                df = self._query_on_executor(query, self.SQLOptions.parse(options))
        except Exception:
            log.error("|%s| Query failed!", self.__class__.__name__)
            raise

        log.info("|Spark| DataFrame successfully created from SQL statement")
        return df

    @slot
    def read_source_as_df(
        self,
        source: str,
        columns: list[str] | None = None,
        hint: str | None = None,
        where: str | None = None,
        df_schema: StructType | None = None,
        window: Window | None = None,
        limit: int | None = None,
        options: JDBCReadOptions | None = None,
    ) -> DataFrame:
        read_options = self._set_lower_upper_bound(
            table=source,
            where=where,
            hint=hint,
            options=self.ReadOptions.parse(options),
        )

        new_columns = columns or ["*"]
        alias: str | None = None

        if read_options.partition_column:
            if read_options.partitioning_mode == JDBCPartitioningMode.MOD:
                partition_column = self.dialect.get_partition_column_mod(
                    read_options.partition_column,
                    read_options.num_partitions,
                )
            elif read_options.partitioning_mode == JDBCPartitioningMode.HASH:
                partition_column = self.dialect.get_partition_column_hash(
                    read_options.partition_column,
                    read_options.num_partitions,
                )
            else:
                partition_column = read_options.partition_column

            # hack to avoid column name verification
            # in the spark, the expression in the partitioning of the column must
            # have the same name as the field in the table ( 2.4 version )
            # https://github.com/apache/spark/pull/21379
            alias = "generated_" + secrets.token_hex(5)
            alias_escaped = self.dialect.escape_column(alias)
            aliased_column = self.dialect.aliased(partition_column, alias_escaped)
            read_options = read_options.copy(update={"partition_column": alias_escaped})
            new_columns.append(aliased_column)

        where = self.dialect.apply_window(where, window)
        query = self.dialect.get_sql_query(
            table=source,
            columns=new_columns,
            where=where,
            hint=hint,
            limit=limit,
        )

        log.info("|%s| Detected dialect: '%s'", self.__class__.__name__, self._get_spark_dialect_name())
        log.info("|%s| Executing SQL query (on executor):", self.__class__.__name__)
        log_lines(log, query)

        result = self._query_on_executor(query, self.ReadOptions.parse(read_options))

        log.info("|Spark| DataFrame successfully created from SQL statement ")
        if alias:
            result = result.drop(alias)

        return result

    @slot
    def write_df_to_target(
        self,
        df: DataFrame,
        target: str,
        options: JDBCWriteOptions | None = None,
    ) -> None:
        write_options = self.WriteOptions.parse(options)
        jdbc_properties = self._get_jdbc_properties(write_options, exclude={"if_exists"}, exclude_none=True)

        mode = (
            "overwrite"
            if write_options.if_exists == JDBCTableExistBehavior.REPLACE_ENTIRE_TABLE
            else write_options.if_exists.value
        )
        log.info("|%s| Saving data to a table %r", self.__class__.__name__, target)
        df.write.format("jdbc").mode(mode).options(dbtable=target, **jdbc_properties).save()
        log.info("|%s| Table %r successfully written", self.__class__.__name__, target)

    @slot
    def get_df_schema(
        self,
        source: str,
        columns: list[str] | None = None,
        options: JDBCReadOptions | None = None,
    ) -> StructType:
        log.info("|%s| Detected dialect: '%s'", self.__class__.__name__, self._get_spark_dialect_name())
        log.info("|%s| Fetching schema of table %r ...", self.__class__.__name__, source)

        query = self.dialect.get_sql_query(source, columns=columns, limit=0, compact=True)
        read_options = self._exclude_partition_options(self.ReadOptions.parse(options), fetchsize=0)

        log.debug("|%s| Executing SQL query (on driver):", self.__class__.__name__)
        log_lines(log, query, level=logging.DEBUG)

        df = self._query_on_driver(query, read_options)
        log.info("|%s| Schema fetched.", self.__class__.__name__)

        return df.schema

    @slot
    def get_min_max_values(
        self,
        source: str,
        window: Window,
        hint: Any | None = None,
        where: Any | None = None,
        options: JDBCReadOptions | None = None,
    ) -> tuple[Any, Any]:
        log.info("|%s| Getting min and max values for expression %r ...", self.__class__.__name__, window.expression)
        read_options = self._exclude_partition_options(self.ReadOptions.parse(options), fetchsize=1)

        query = self.dialect.get_sql_query(
            table=source,
            columns=[
                self.dialect.aliased(
                    self.dialect.get_min_value(window.expression),
                    self.dialect.escape_column("min"),
                ),
                self.dialect.aliased(
                    self.dialect.get_max_value(window.expression),
                    self.dialect.escape_column("max"),
                ),
            ],
            where=self.dialect.apply_window(where, window),
            hint=hint,
        )

        log.info("|%s| Executing SQL query (on driver):", self.__class__.__name__)
        log_lines(log, query)

        df = self._query_on_driver(query, read_options)
        row = df.collect()[0]
        min_value = row["min"]
        max_value = row["max"]

        log.info("|%s| Received values:", self.__class__.__name__)
        log_with_indent(log, "MIN(%s) = %r", window.expression, min_value)
        log_with_indent(log, "MAX(%s) = %r", window.expression, max_value)

        return min_value, max_value

    def _query_on_executor(
        self,
        query: str,
        options: JDBCSQLOptions | JDBCReadOptions,
    ) -> DataFrame:
        jdbc_properties = self._get_jdbc_properties(options, exclude_none=True)
        return self.spark.read.format("jdbc").options(dbtable=f"({query}) T", **jdbc_properties).load()

    def _exclude_partition_options(
        self,
        options: JDBCReadOptions,
        fetchsize: int,
    ) -> JDBCFetchOptions:
        return self.FetchOptions.parse(
            options.copy(
                update={"fetchsize": fetchsize},
                exclude={"partition_column", "lower_bound", "upper_bound", "num_partitions", "partitioning_mode"},
            ).dict(),
        )

    def _set_lower_upper_bound(
        self,
        table: str,
        hint: str | None,
        where: str | None,
        options: JDBCReadOptions,
    ) -> JDBCReadOptions:
        """
        Determine values of upperBound and lowerBound options
        """
        if not options.partition_column:
            return options

        missing_values: list[str] = []

        is_missed_lower_bound = options.lower_bound is None
        is_missed_upper_bound = options.upper_bound is None

        if is_missed_lower_bound:
            missing_values.append("lowerBound")

        if is_missed_upper_bound:
            missing_values.append("upperBound")

        if not missing_values:
            return options

        log.warning(
            "|%s| Passed numPartitions = %d, but values %r are not set. "
            "They will be detected automatically based on values in partitionColumn %r",
            self.__class__.__name__,
            options.num_partitions,
            missing_values,
            options.partition_column,
        )

        min_partition_value, max_partition_value = self.get_min_max_values(
            source=table,
            window=Window(options.partition_column),
            where=where,
            hint=hint,
            options=options,
        )

        # The sessionInitStatement parameter is removed because it only needs to be applied once.
        return options.copy(
            exclude={"session_init_statement"},
            update={
                "lower_bound": options.lower_bound if not is_missed_lower_bound else min_partition_value,
                "upper_bound": options.upper_bound if not is_missed_upper_bound else max_partition_value,
            },
        )

    def _log_parameters(self):
        super()._log_parameters()
        log_with_indent(log, "jdbc_url = %r", self.jdbc_url)



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from typing import TYPE_CHECKING

try:
    from pydantic.v1 import Field, SecretStr
except (ImportError, AttributeError):
    from pydantic import Field, SecretStr  # type: ignore[no-redef, assignment]

from onetl.connection.db_connection.kafka.kafka_auth import KafkaAuth
from onetl.impl import GenericOptions

if TYPE_CHECKING:
    from onetl.connection import Kafka


class KafkaBasicAuth(KafkaAuth, GenericOptions):
    """
    Connect to Kafka using ``sasl.mechanism="PLAIN"``.

    For more details see `Kafka Documentation <https://kafka.apache.org/documentation/#security_sasl_plain>`_.

    .. versionadded:: 0.9.0

    Examples
    --------

    Auth in Kafka with user and password:

    .. code:: python

        from onetl.connection import Kafka

        auth = Kafka.BasicAuth(
            user="some_user",
            password="abc",
        )

    """

    user: str = Field(alias="username")
    password: SecretStr

    def get_jaas_conf(self) -> str:  # noqa: WPS473
        return (
            "org.apache.kafka.common.security.plain.PlainLoginModule required "
            f'username="{self.user}" '
            f'password="{self.password.get_secret_value()}";'
        )

    def get_options(self, kafka: Kafka) -> dict:
        return {
            "sasl.mechanism": "PLAIN",
            "sasl.jaas.config": self.get_jaas_conf(),
        }

    def cleanup(self, kafka: Kafka) -> None:
        # nothing to cleanup
        pass



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import os
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, ContextManager

from onetl.base.base_connection import BaseConnection
from onetl.base.base_file_format import BaseReadableFileFormat, BaseWritableFileFormat
from onetl.base.pure_path_protocol import PurePathProtocol

if TYPE_CHECKING:
    from pyspark.sql import DataFrame, DataFrameReader, DataFrameWriter, SparkSession
    from pyspark.sql.types import StructType


class FileDFReadOptions(ABC):
    """
    Protocol for objects supporting altering Spark DataFrameReader options.

    .. versionadded:: 0.9.0
    """

    @abstractmethod
    def apply_to_reader(self, reader: DataFrameReader) -> DataFrameReader | ContextManager[DataFrameReader]:
        """
        Apply provided format to :obj:`pyspark.sql.DataFrameReader`.

        .. versionadded:: 0.9.0

        Returns
        -------
        :obj:`pyspark.sql.DataFrameReader`
            DataFrameReader with options applied.

        ``ContextManager[DataFrameReader]``
            If returned context manager, it will be entered before reading data and exited after creating a DataFrame.
            Context manager's ``__enter__`` method should return :obj:`pyspark.sql.DataFrameReader` instance.
        """


class FileDFWriteOptions(ABC):
    """
    Protocol for objects supporting altering Spark DataFrameWriter options.

    .. versionadded:: 0.9.0
    """

    @abstractmethod
    def apply_to_writer(self, writer: DataFrameWriter) -> DataFrameWriter | ContextManager[DataFrameWriter]:
        """
        Apply provided format to :obj:`pyspark.sql.DataFrameWriter`.

        .. versionadded:: 0.9.0

        Returns
        -------
        :obj:`pyspark.sql.DataFrameWriter`
            DataFrameWriter with options applied.

        ``ContextManager[DataFrameWriter]``
            If returned context manager, it will be entered before writing and exited after writing a DataFrame.
            Context manager's ``__enter__`` method should return :obj:`pyspark.sql.DataFrameWriter` instance.
        """


class BaseFileDFConnection(BaseConnection):
    """
    Implements generic methods for reading  and writing dataframe as files.

    .. versionadded:: 0.9.0
    """

    spark: SparkSession

    @abstractmethod
    def check_if_format_supported(
        self,
        format: BaseReadableFileFormat | BaseWritableFileFormat,  # noqa: WPS125
    ) -> None:
        """
        Validate if specific file format is supported. |support_hooks|

        .. versionadded:: 0.9.0

        Raises
        ------
        RuntimeError
            If file format is not supported.
        """

    @abstractmethod
    def path_from_string(self, path: os.PathLike | str) -> PurePathProtocol:
        """
        Convert path from string to object. |support_hooks|

        .. versionadded:: 0.9.0
        """

    @property
    @abstractmethod
    def instance_url(self) -> str:
        """Instance URL.

        .. versionadded:: 0.9.0
        """

    @abstractmethod
    def read_files_as_df(
        self,
        paths: list[PurePathProtocol],
        format: BaseReadableFileFormat,  # noqa: WPS125
        root: PurePathProtocol | None = None,
        df_schema: StructType | None = None,
        options: FileDFReadOptions | None = None,
    ) -> DataFrame:
        """
        Read files in some paths list as dataframe. |support_hooks|

        .. versionadded:: 0.9.0
        """

    @abstractmethod
    def write_df_as_files(
        self,
        df: DataFrame,
        path: PurePathProtocol,
        format: BaseWritableFileFormat,  # noqa: WPS125
        options: FileDFWriteOptions | None = None,
    ) -> None:
        """
        Write dataframe as files in some path. |support_hooks|

        .. versionadded:: 0.9.0
        """



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
import ftplib  # noqa: S402  # nosec
import textwrap

from ftputil import FTPHost
from ftputil import session as ftp_session

try:
    from onetl.connection.file_connection.ftp import FTP
except (ImportError, NameError) as e:
    raise ImportError(
        textwrap.dedent(
            """
            Cannot import module "ftputil".

            Since onETL v0.7.0 you should install package as follows:
                pip install onetl[ftps]

            or
                pip install onetl[files]
            """,
        ).strip(),
    ) from e


class TLSfix(ftplib.FTP_TLS):  # noqa: N801
    """
    Fix for python 3.6+
    https://stackoverflow.com/questions/14659154/ftpes-session-reuse-required
    """

    def ntransfercmd(self, cmd, rest=None):
        conn, size = ftplib.FTP.ntransfercmd(self, cmd, rest)  # noqa: S321  # nosec
        if self._prot_p:
            conn = self.context.wrap_socket(
                conn,
                server_hostname=self.host,
                session=self.sock.session,
            )  # this is the fix
        return conn, size


class FTPS(FTP):
    """FTPS file connection. |support_hooks|

    Based on `FTPUtil library <https://pypi.org/project/ftputil/>`_.

    .. warning::

        Since onETL v0.7.0 to use FTPS connector you should install package as follows:

        .. code:: bash

            pip install onetl[ftps]

            # or
            pip install onetl[files]

        See :ref:`install-files` installation instruction for more details.

    .. versionadded:: 0.1.0

    Parameters
    ----------
    host : str
        Host of FTPS source. For example: ``ftps.domain.com``

    port : int, default: ``21``
        Port of FTPS source

    user : str, default: ``None``
        User, which have access to the file source. For example: ``someuser``.

        ``None`` means that the user is anonymous.

    password : str, default: ``None``
        Password for file source connection.

        ``None`` means that the user is anonymous.

    Examples
    --------

    FTPS file connection initialization

    .. code:: python

        from onetl.connection import FTPS

        ftps = FTPS(
            host="ftps.domain.com",
            user="someuser",
            password="*****",
        )
    """

    def _get_client(self) -> FTPHost:
        """
        Returns a FTPS connection object
        """

        session_factory = ftp_session.session_factory(
            base_class=TLSfix,
            port=self.port,
            encrypt_data_channel=True,
            debug_level=0,
        )

        return FTPHost(
            self.host,
            self.user,
            self.password.get_secret_value() if self.password else None,
            session_factory=session_factory,
        )



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import logging
from typing import ClassVar, Optional

from etl_entities.instance import Host

from onetl._util.classproperty import classproperty
from onetl._util.version import Version
from onetl.connection.db_connection.clickhouse.dialect import ClickhouseDialect
from onetl.connection.db_connection.clickhouse.options import (
    ClickhouseExecuteOptions,
    ClickhouseFetchOptions,
    ClickhouseReadOptions,
    ClickhouseSQLOptions,
    ClickhouseWriteOptions,
)
from onetl.connection.db_connection.jdbc_connection import JDBCConnection
from onetl.connection.db_connection.jdbc_mixin import JDBCStatementType
from onetl.hooks import slot, support_hooks
from onetl.impl import GenericOptions

# do not import PySpark here, as we allow user to use `Clickhouse.get_packages()` for creating Spark session


log = logging.getLogger(__name__)


class ClickhouseExtra(GenericOptions):
    class Config:
        extra = "allow"


@support_hooks
class Clickhouse(JDBCConnection):
    """Clickhouse JDBC connection. |support_hooks|

    Based on Maven package `com.clickhouse:clickhouse-jdbc:0.6.5 <https://mvnrepository.com/artifact/com.clickhouse/clickhouse-jdbc/0.6.5>`_
    (`official Clickhouse JDBC driver <https://github.com/ClickHouse/clickhouse-jdbc>`_).

    .. seealso::

        Before using this connector please take into account :ref:`clickhouse-prerequisites`

    .. versionadded:: 0.1.0

    Parameters
    ----------
    host : str
        Host of Clickhouse database. For example: ``test.clickhouse.domain.com`` or ``193.168.1.11``

    port : int, default: ``8123``
        Port of Clickhouse database

    user : str
        User, which have proper access to the database. For example: ``some_user``

    password : str
        Password for database connection

    database : str, optional
        Database in RDBMS, NOT schema.

        See `this page <https://www.educba.com/postgresql-database-vs-schema/>`_ for more details

    spark : :obj:`pyspark.sql.SparkSession`
        Spark session.

    extra : dict, default: ``None``
        Specifies one or more extra parameters by which clients can connect to the instance.

        For example: ``{"continueBatchOnError": "false"}``.

        See:
            * `Clickhouse JDBC driver properties documentation <https://clickhouse.com/docs/en/integrations/java#configuration>`_
            * `Clickhouse core settings documentation <https://clickhouse.com/docs/en/operations/settings/settings>`_
            * `Clickhouse query complexity documentation <https://clickhouse.com/docs/en/operations/settings/query-complexity>`_
            * `Clickhouse query level settings <https://clickhouse.com/docs/en/operations/settings/query-level>`_

    Examples
    --------

    Clickhouse connection initialization

    .. code:: python

        from onetl.connection import Clickhouse
        from pyspark.sql import SparkSession

        # Create Spark session with Clickhouse driver loaded
        maven_packages = Clickhouse.get_packages()
        spark = (
            SparkSession.builder.appName("spark-app-name")
            .config("spark.jars.packages", ",".join(maven_packages))
            .getOrCreate()
        )

        # Create connection
        clickhouse = Clickhouse(
            host="database.host.or.ip",
            user="user",
            password="*****",
            extra={"continueBatchOnError": "false"},
            spark=spark,
        )

    """

    host: Host
    port: int = 8123
    database: Optional[str] = None
    extra: ClickhouseExtra = ClickhouseExtra()

    Extra = ClickhouseExtra
    Dialect = ClickhouseDialect

    ReadOptions = ClickhouseReadOptions
    WriteOptions = ClickhouseWriteOptions
    SQLOptions = ClickhouseSQLOptions
    FetchOptions = ClickhouseFetchOptions
    ExecuteOptions = ClickhouseExecuteOptions

    DRIVER: ClassVar[str] = "com.clickhouse.jdbc.ClickHouseDriver"

    @slot
    @classmethod
    def get_packages(
        cls,
        package_version: str | None = None,
        apache_http_client_version: str | None = None,
    ) -> list[str]:
        """
        Get package names to be downloaded by Spark. Allows specifying custom JDBC and Apache HTTP Client versions. |support_hooks|

        .. versionadded:: 0.9.0

        Parameters
        ----------
        package_version : str, optional
            ClickHouse JDBC version client packages. Defaults to ``0.6.5``.

            .. versionadded:: 0.11.0

        apache_http_client_version : str, optional
            Apache HTTP Client version package. Defaults to ``5.3.1``.

            .. versionadded:: 0.11.0

        Examples
        --------

        .. code:: python

            from onetl.connection import Clickhouse

            Clickhouse.get_packages(package_version="0.6.0", apache_http_client_version="5.3.1")

        """
        default_jdbc_version = "0.6.5"
        default_http_version = "5.3.1"

        jdbc_version = Version(package_version or default_jdbc_version).min_digits(3)
        http_version = Version(apache_http_client_version or default_http_version).min_digits(3)

        result = [
            f"com.clickhouse:clickhouse-jdbc:{jdbc_version}",
            f"com.clickhouse:clickhouse-http-client:{jdbc_version}",
        ]

        if jdbc_version >= Version("0.5.0"):
            result.append(f"org.apache.httpcomponents.client5:httpclient5:{http_version}")

        return result

    @classproperty
    def package(self) -> str:
        """Get a single string of package names to be downloaded by Spark for establishing a Clickhouse connection."""
        return "com.clickhouse:clickhouse-jdbc:0.6.5,com.clickhouse:clickhouse-http-client:0.6.5,org.apache.httpcomponents.client5:httpclient5:5.3.1"

    @property
    def jdbc_url(self) -> str:
        if self.database:
            return f"jdbc:clickhouse://{self.host}:{self.port}/{self.database}"

        return f"jdbc:clickhouse://{self.host}:{self.port}"

    @property
    def jdbc_params(self) -> dict:
        result = super().jdbc_params
        result.update(self.extra.dict(by_alias=True))
        return result

    @property
    def instance_url(self) -> str:
        return f"{self.__class__.__name__.lower()}://{self.host}:{self.port}"

    def __str__(self):
        return f"{self.__class__.__name__}[{self.host}:{self.port}]"

    @staticmethod
    def _build_statement(
        statement: str,
        statement_type: JDBCStatementType,
        jdbc_connection,
        statement_args,
    ):
        # Clickhouse does not support prepared statements, as well as calling functions/procedures
        return jdbc_connection.createStatement(*statement_args)



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from onetl._metrics.listener.execution import (
    SparkListenerExecution,
    SparkListenerExecutionStatus,
    SparkSQLMetricNames,
)
from onetl._metrics.listener.job import SparkListenerJob, SparkListenerJobStatus
from onetl._metrics.listener.listener import SparkMetricsListener
from onetl._metrics.listener.stage import SparkListenerStage, SparkListenerStageStatus
from onetl._metrics.listener.task import (
    SparkListenerTask,
    SparkListenerTaskMetrics,
    SparkListenerTaskStatus,
)

__all__ = [
    "SparkListenerTask",
    "SparkListenerTaskStatus",
    "SparkListenerTaskMetrics",
    "SparkListenerStage",
    "SparkListenerStageStatus",
    "SparkListenerJob",
    "SparkListenerJobStatus",
    "SparkListenerExecution",
    "SparkListenerExecutionStatus",
    "SparkSQLMetricNames",
    "SparkMetricsListener",
]



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import os
from abc import abstractmethod
from typing import Iterable

from onetl.base.base_connection import BaseConnection
from onetl.base.base_file_filter import BaseFileFilter
from onetl.base.base_file_limit import BaseFileLimit
from onetl.base.path_protocol import PathWithStatsProtocol
from onetl.base.path_stat_protocol import PathStatProtocol


class BaseFileConnection(BaseConnection):
    """
    Implements generic methods for files and directories manipulation on some filesystem (usually remote).

    .. versionadded:: 0.8.0
    """

    @abstractmethod
    def path_exists(self, path: os.PathLike | str) -> bool:
        """
        Check if specified path exists on remote filesystem. |support_hooks|.

        .. versionadded:: 0.8.0

        Parameters
        ----------
        path : str or :obj:`os.PathLike`
            Path to check

        Returns
        -------
        ``True`` if path exists, ``False`` otherwise

        Examples
        --------

        >>> connection.path_exists("/path/to/file.csv")
        True
        >>> connection.path_exists("/path/to/dir")
        True
        >>> connection.path_exists("/path/to/missing")
        False
        """

    @abstractmethod
    def is_file(self, path: os.PathLike | str) -> bool:
        """
        Check if specified path is a file. |support_hooks|

        .. versionadded:: 0.8.0

        Parameters
        ----------
        path : str or :obj:`os.PathLike`
            Path to check

        Returns
        -------
        ``True`` if path is a file, ``False`` otherwise.

        Raises
        ------
        FileNotFoundError
            Path does not exist

        Examples
        --------

        >>> connection.is_file("/path/to/dir/file.csv")
        True
        >>> connection.is_file("/path/to/dir")
        False
        """

    @abstractmethod
    def is_dir(self, path: os.PathLike | str) -> bool:
        """
        Check if specified path is a directory. |support_hooks|

        .. versionadded:: 0.8.0

        Parameters
        ----------
        path : str or :obj:`os.PathLike`
            Path to check

        Returns
        -------
        ``True`` if path is a directory, ``False`` otherwise.

        Raises
        ------
        :obj:`onetl.exception.DirectoryNotFoundError`
            Path does not exist

        Examples
        --------

        >>> connection.is_dir("/path/to/dir")
        True
        >>> connection.is_dir("/path/to/dir/file.csv")
        False
        """

    @abstractmethod
    def get_stat(self, path: os.PathLike | str) -> PathStatProtocol:
        """
        Returns stats for a specific path. |support_hooks|

        .. versionadded:: 0.8.0

        Parameters
        ----------
        path : str or :obj:`os.PathLike`
            Path to get stats for

        Returns
        -------
        Stats object

        Raises
        ------
        Any underlying client exception

        Examples
        --------

        >>> stat = connection.get_stat("/path/to/file.csv")
        >>> stat.st_size  # in bytes
        1024
        >>> stat.st_uid  # owner id or name
        12345
        """

    @abstractmethod
    def resolve_dir(self, path: os.PathLike | str) -> PathWithStatsProtocol:
        """
        Returns directory at specific path, with stats. |support_hooks|

        .. versionadded:: 0.8.0

        Parameters
        ----------
        path : str or :obj:`os.PathLike`
            Path to resolve

        Returns
        -------
        Directory path with stats

        Raises
        ------
        :obj:`onetl.exception.DirectoryNotFoundError`
            Path does not exist

        NotADirectoryError
            Path is not a directory

        Examples
        --------

        >>> dir_path = connection.resolve_dir("/path/to/dir")
        >>> os.fspath(dir_path)
        '/path/to/dir'
        >>> dir_path.stat.st_uid  # owner id
        12345
        """

    @abstractmethod
    def resolve_file(self, path: os.PathLike | str) -> PathWithStatsProtocol:
        """
        Returns file at specific path, with stats. |support_hooks|

        .. versionadded:: 0.8.0

        Parameters
        ----------
        path : str or :obj:`os.PathLike`
            Path to resolve

        Returns
        -------
        File path with stats

        Raises
        ------
        FileNotFoundError
            Path does not exist

        :obj:`onetl.exception.NotAFileError`
            Path is not a file

        Examples
        --------

        >>> file_path = connection.resolve_file("/path/to/dir/file.csv")
        >>> os.fspath(file_path)
        '/path/to/dir/file.csv'
        >>> file_path.stat.st_uid  # owner id
        12345
        """

    @abstractmethod
    def create_dir(self, path: os.PathLike | str) -> PathWithStatsProtocol:
        """
        Creates directory tree on remote filesystem. |support_hooks|

        .. versionadded:: 0.8.0

        Parameters
        ----------
        path : str or :obj:`os.PathLike`
            Directory path

        Returns
        -------
        Created directory with stats

        Raises
        ------
        :obj:`onetl.exception.NotAFileError`
            Path is not a file

        Examples
        --------

        >>> dir_path = connection.create_dir("/path/to/dir")
        >>> os.fspath(dir_path)
        '/path/to/dir'
        """

    @abstractmethod
    def remove_file(self, path: os.PathLike | str) -> bool:
        """
        Removes file on remote filesystem. |support_hooks|

        If file does not exist, no exception is raised.

        .. warning::

            Supports only one file removal per call. Directory removal is **NOT** supported, use :obj:`~remove_dir` instead.

        .. versionadded:: 0.8.0

        Parameters
        ----------
        path : str or :obj:`os.PathLike`
            File path

        Returns
        -------
        ``True`` if file was removed, ``False`` if file does not exist in the first place.

        Raises
        ------
        :obj:`onetl.exception.NotAFileError`
            Path is not a file

        Examples
        --------

        >>> connection.remove_file("/path/to/file.csv")
        True
        >>> connection.path_exists("/path/to/dir/file.csv")
        False
        >>> connection.remove_file("/path/to/file.csv")  # already deleted, no error
        False
        """

    @abstractmethod
    def remove_dir(self, path: os.PathLike | str, recursive: bool = False) -> bool:
        """
        Remove directory or directory tree. |support_hooks|

        If directory does not exist, no exception is raised.

        .. versionadded:: 0.8.0

        Parameters
        ----------
        path : str or :obj:`os.PathLike`
            Directory path to remote

        recursive : bool, default ``False``
            If ``True``, remove directory tree recursively.

        Returns
        -------
        ``True`` if directory was removed, ``False`` if directory does not exist in the first place.

        Raises
        ------
        NotADirectoryError
            Path is not a directory

        Examples
        --------

        >>> connection.remove_dir("/path/to/dir")
        True
        >>> connection.path_exists("/path/to/dir")
        False
        >>> connection.path_exists("/path/to/dir/file.csv")
        False
        >>> connection.remove_dir("/path/to/dir")  # already deleted, no error
        False
        """

    @abstractmethod
    def rename_file(
        self,
        source_file_path: os.PathLike | str,
        target_file_path: os.PathLike | str,
        replace: bool = False,
    ) -> PathWithStatsProtocol:
        """
        Rename or move file on remote filesystem. |support_hooks|

        .. warning::

            Supports only one file move per call. Directory move/rename is **NOT** supported.

        .. versionadded:: 0.8.0

        Parameters
        ----------
        source_file_path : str or :obj:`os.PathLike`
            Old file path

        target_file_path : str or :obj:`os.PathLike`
            New file path

        replace : bool, default ``False``
            If ``True``, existing file will be replaced.

        Returns
        -------
        New file path with stats.

        Raises
        ------
        :obj:`onetl.exception.NotAFileError`
            Source or target path is not a file

        FileNotFoundError
            File does not exist

        FileExistsError
            File already exists, and ``replace=False``

        Examples
        --------

        >>> new_file = connection.rename_file("/path/to/file1.csv", "/path/to/file2.csv")
        >>> os.fspath(new_file)
        '/path/to/file2.csv'
        >>> connection.path_exists("/path/to/file2.csv")
        True
        >>> connection.path_exists("/path/to/file1.csv")
        False
        """

    @abstractmethod
    def list_dir(
        self,
        path: os.PathLike | str,
        filters: Iterable[BaseFileFilter] | None = None,
        limits: Iterable[BaseFileLimit] | None = None,
    ) -> list[PathWithStatsProtocol]:
        """
        Return list of child files/directories in a specific directory. |support_hooks|

        .. versionadded:: 0.8.0

        Parameters
        ----------
        path : str or :obj:`os.PathLike`
            Directory path to list contents.

        filters : list of :obj:`BaseFileFilter <onetl.base.base_file_filter.BaseFileFilter>`, optional
            Return only files/directories matching these filters. See :ref:`file-filters`

        limits : list of :obj:`BaseFileLimit <onetl.base.base_file_limit.BaseFileLimit>`, optional
            Apply limits to the list of files/directories, and stop if one of the limits is reached.
            See :ref:`file-limits`

        Returns
        -------
        List of :obj:`onetl.base.PathWithStatsProtocol`

        Raises
        ------
        NotADirectoryError
            Path is not a directory

        :obj:`onetl.exception.DirectoryNotFoundError`
            Path does not exist

        Examples
        --------

        >>> dir_content = connection.list_dir("/path/to/dir")
        >>> os.fspath(dir_content[0])
        '/path/to/dir/file.csv'
        >>> connection.path_exists("/path/to/dir/file.csv")
        True
        """

    @abstractmethod
    def walk(
        self,
        root: os.PathLike | str,
        topdown: bool = True,
        filters: Iterable[BaseFileFilter] | None = None,
        limits: Iterable[BaseFileLimit] | None = None,
    ) -> Iterable[tuple[PathWithStatsProtocol, list[PathWithStatsProtocol], list[PathWithStatsProtocol]]]:
        """
        Walk into directory tree, and iterate over its content in all nesting levels. |support_hooks|

        Just like :obj:`os.walk`, but with additional filter/limit logic.

        .. versionadded:: 0.8.0

        Parameters
        ----------
        root : str or :obj:`os.PathLike`
            Directory path to walk into.

        topdown : bool, default ``True``
            If ``True``, walk in top-down order, otherwise walk in bottom-up order.

        filters : list of :obj:`BaseFileFilter <onetl.base.base_file_filter.BaseFileFilter>`, optional
            Return only files/directories matching these filters. See :ref:`file-filters`

        limits : list of :obj:`BaseFileLimit <onetl.base.base_file_limit.BaseFileLimit>`, optional
            Apply limits to the list of files/directories, and stop if one of the limits is reached.
            See :ref:`file-limits`

        Returns
        -------
        ``Iterator[tuple[root, dirs, files]]``, like :obj:`os.walk`.

        But all the paths are not strings, instead path classes with embedded stats are returned.

        Raises
        ------
        NotADirectoryError
            Path is not a directory

        :obj:`onetl.exception.DirectoryNotFoundError`
            Path does not exist

        Examples
        --------

        >>> for root, dirs, files in connection.walk("/path/to/dir"):
        ...    break
        >>> os.fspath(root)
        '/path/to/dir'
        >>> dirs
        []
        >>> os.fspath(files[0])
        '/path/to/dir/file.csv'
        >>> connection.path_exists("/path/to/dir/file.csv")
        True
        """

    @abstractmethod
    def download_file(
        self,
        remote_file_path: os.PathLike | str,
        local_file_path: os.PathLike | str,
        replace: bool = True,
    ) -> PathWithStatsProtocol:
        """
        Downloads file from the remote filesystem to a local path. |support_hooks|

        .. warning::

            Supports only one file download per call. Directory download is **NOT** supported, use :ref:`file-downloader` instead.

        .. versionadded:: 0.8.0

        Parameters
        ----------
        remote_file_path : str or :obj:`os.PathLike`
            Remote file path to read from

        local_file_path : str or :obj:`os.PathLike`
            Local file path to create

        replace : bool, default ``False``
            If ``True``, existing file will be replaced

        Returns
        -------
        Local file with stats.

        Raises
        ------
        :obj:`onetl.exception.NotAFileError`
            Remote or local path is not a file

        FileNotFoundError
            Remote file does not exist

        FileExistsError
            Local file already exists, and ``replace=False``

        :obj:`onetl.exception.FileSizeMismatchError`
            Target file size after download is different from source file size.

        Examples
        --------

        >>> local_file = connection.download_file(
        ...     remote_file_path="/path/to/source.csv",
        ...     local_file_path="/path/to/target.csv",
        ... )
        >>> os.fspath(local_file)
        '/path/to/target.csv'
        >>> local_file.exists()
        True
        >>> local_file.stat().st_size  # in bytes
        1024
        >>> connection.get_stat("/path/to/source.csv").st_size  # same size
        1024
        """

    @abstractmethod
    def upload_file(
        self,
        local_file_path: os.PathLike | str,
        remote_file_path: os.PathLike | str,
        replace: bool = False,
    ) -> PathWithStatsProtocol:
        """
        Uploads local file to a remote filesystem. |support_hooks|

        .. warning::

            Supports only one file upload per call. Directory upload is **NOT** supported, use :ref:`file-uploader` instead.

        .. versionadded:: 0.8.0

        Parameters
        ----------
        local_file_path : str or :obj:`os.PathLike`
            Local file path to read from

        remote_file_path : str or :obj:`os.PathLike`
            Remote file path to create

        replace : bool, default ``False``
            If ``True``, existing file will be replaced

        Returns
        -------
        Remote file with stats.

        Raises
        ------
        :obj:`onetl.exception.NotAFileError`
            Remote or local path is not a file

        FileNotFoundError
            Local file does not exist

        FileExistsError
            Remote file already exists, and ``replace=False``

        :obj:`onetl.exception.FileSizeMismatchError`
            Target file size after upload is different from source file size.

        Examples
        --------

        >>> remote_file = connection.upload(
        ...     local_file_path="/path/to/source.csv",
        ...     remote_file_path="/path/to/target.csv",
        ... )
        >>> os.fspath(remote_file)
        '/path/to/target.csv'
        >>> connection.path_exists("/path/to/target.csv")
        True
        >>> remote_file.stat().st_size  # in bytes
        1024
        >>> os.stat("/path/to/source.csv").st_size  # same as source
        1024
        """

    @abstractmethod
    def read_text(self, path: os.PathLike | str, encoding: str = "utf-8") -> str:
        r"""
        Returns string content of a file at specific path. |support_hooks|

        .. versionadded:: 0.8.0

        Parameters
        ----------
        path : str or :obj:`os.PathLike`
            File path to read

        encoding : str, default ``utf-8``
            File content encoding

        Returns
        -------
        File content

        Raises
        ------
        FileNotFoundError
            Path does not exist

        :obj:`onetl.exception.NotAFileError`
            Path is not a file

        Examples
        --------

        >>> connection.read_text("/path/to/dir/file.csv")
        'some;header\n1;2'
        """

    @abstractmethod
    def read_bytes(self, path: os.PathLike | str) -> bytes:
        """
        Returns binary content of a file at specific path. |support_hooks|

        .. versionadded:: 0.8.0

        Parameters
        ----------
        path : str or :obj:`os.PathLike`
            File path to read

        Returns
        -------
        File content

        Raises
        ------
        FileNotFoundError
            Path does not exist

        :obj:`onetl.exception.NotAFileError`
            Path is not a file

        Examples
        --------

        >>> connection.read_bytes("/path/to/dir/file.csv")
        b'0xdeadbeef'
        """

    @abstractmethod
    def write_text(
        self,
        path: os.PathLike | str,
        content: str,
        encoding: str = "utf-8",
    ) -> PathWithStatsProtocol:
        r"""
        Writes string content to a file at specific path. |support_hooks|

        .. warning::

            If file already exists, its content will be replaced.

        .. versionadded:: 0.8.0

        Parameters
        ----------
        path : str or :obj:`os.PathLike`
            File path to write

        content : str
            File content

        encoding : str, default ``utf-8``
            File content encoding

        Returns
        -------
        File path with stats after write

        Raises
        ------
        TypeError
            Content is not a string

        :obj:`onetl.exception.NotAFileError`
            Path is not a file

        Examples
        --------

        >>> file_path = connection.write_text("/path/to/dir/file.csv", "some;header\n1;2")
        >>> os.fspath(file_path)
        '/path/to/dir/file.csv'
        >>> file_path.stat.st_size  # in bytes
        1024
        """

    @abstractmethod
    def write_bytes(self, path: os.PathLike | str, content: bytes) -> PathWithStatsProtocol:
        """
        Writes bytes content to a file at specific path. |support_hooks|

        .. warning::

            If file already exists, its content will be replaced.

        .. versionadded:: 0.8.0

        Parameters
        ----------
        path : str or :obj:`os.PathLike`
            File path to write

        content : bytes
            File content

        Returns
        -------
        File path with stats after write

        Raises
        ------
        TypeError
            Content is not a string

        :obj:`onetl.exception.NotAFileError`
            Path is not a file

        Examples
        --------

        >>> file_path = connection.write_bytes("/path/to/dir/file.csv", b"0xdeadbeef")
        >>> os.fspath(file_path)
        '/path/to/dir/file.csv'
        >>> file_path.stat.st_size  # in bytes
        1024
        """

    @property
    @abstractmethod
    def instance_url(self):
        """Instance URL"""



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from onetl.connection.file_df_connection.spark_hdfs.connection import SparkHDFS
from onetl.connection.file_df_connection.spark_hdfs.slots import SparkHDFSSlots



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from datetime import datetime
from typing import Any, Iterable, Mapping

from onetl.connection.db_connection.db_connection.dialect import DBDialect
from onetl.connection.db_connection.dialect_mixins import (
    NotSupportColumns,
    RequiresDFSchema,
    SupportHWMExpressionStr,
    SupportNameAny,
)
from onetl.hwm import Edge, Window

_upper_level_operators = frozenset(  # noqa: WPS527
    [
        "$addFields",
        "$bucket",
        "$bucketAuto",
        "$changeStream",
        "$collStats",
        "$count",
        "$currentOp",
        "$densify",
        "$documents",
        "$facet",
        "$fill",
        "$geoNear",
        "$graphLookup",
        "$group",
        "$indexStats",
        "$limit",
        "$listLocalSessions",
        "$listSessions",
        "$lookup",
        "$merge",
        "$out",
        "$planCacheStats",
        "$project",
        "$redact",
        "$replaceRoot",
        "$replaceWith",
        "$sample",
        "$search",
        "$searchMeta",
        "$set",
        "$setWindowFields",
        "$shardedDataDistribution",
        "$skip",
        "$sort",
        "$sortByCount",
        "$unionWith",
        "$unset",
        "$unwind",
    ],
)


class MongoDBDialect(  # noqa: WPS215
    SupportNameAny,
    NotSupportColumns,
    RequiresDFSchema,
    SupportHWMExpressionStr,
    DBDialect,
):
    def validate_where(
        self,
        where: Any,
    ) -> dict | None:
        if where is None:
            return None

        if not isinstance(where, dict):
            raise ValueError(
                f"{self.connection.__class__.__name__} requires 'where' parameter type to be 'dict', "
                f"got {where.__class__.__name__!r}",
            )

        for key in where:
            self._validate_top_level_keys_in_where_parameter(key)
        return where

    def validate_hint(
        self,
        hint: Any,
    ) -> dict | None:
        if hint is None:
            return None

        if not isinstance(hint, dict):
            raise ValueError(
                f"{self.connection.__class__.__name__} requires 'hint' parameter type to be 'dict', "
                f"got {hint.__class__.__name__!r}",
            )
        return hint

    def prepare_pipeline(
        self,
        pipeline: Any,
    ) -> Any:
        """
        Prepares pipeline (list or dict) to MongoDB syntax, but without converting it to string.
        """

        if isinstance(pipeline, Mapping):
            return {self.prepare_pipeline(key): self.prepare_pipeline(value) for key, value in pipeline.items()}

        if isinstance(pipeline, Iterable) and not isinstance(pipeline, str):
            return [self.prepare_pipeline(item) for item in pipeline]

        return self._serialize_value(pipeline)

    def apply_window(
        self,
        condition: Any,
        window: Window | None = None,
    ) -> Any:
        result = super().apply_window(condition, window)
        if not result:
            return {}
        if len(result) == 1:
            return result[0]
        return {"$and": result}

    def _serialize_value(self, value: Any) -> str | int | dict:
        """
        Transform the value into an SQL Dialect-supported form.
        """

        if isinstance(value, datetime):
            return {"$date": value.astimezone().isoformat()}

        return value

    def _edge_to_where(
        self,
        expression: str,
        edge: Edge,
        position: str,
    ) -> Any:
        if not expression or not edge.is_set():
            return None

        operators: dict[tuple[str, bool], str] = {
            ("start", True): "$gte",
            ("start", False): "$gt",
            ("end", True): "$lte",
            ("end", False): "$lt",
        }

        operator = operators[(position, edge.including)]
        value = self._serialize_value(edge.value)
        return {
            expression: {
                operator: value,
            },
        }

    def _validate_top_level_keys_in_where_parameter(self, key: str):
        """
        Checks the 'where' parameter for illegal operators, such as ``$match``, ``$merge`` or ``$changeStream``.

        'where' clause can contain only filtering operators, like ``{"col1" {"$eq": 1}}`` or ``{"$and": [...]}``.
        """
        if key.startswith("$"):
            if key == "$match":
                raise ValueError(
                    "'$match' operator not allowed at the top level of the 'where' parameter dictionary. "
                    "This error most likely occurred due to the fact that you used the MongoDB format for the "
                    "pipeline {'$match': {'column': ...}}. In the onETL paradigm, you do not need to specify the "
                    "'$match' keyword, but write the filtering condition right away, like {'column': ...}",
                )
            if key in _upper_level_operators:  # noqa: WPS220
                raise ValueError(  # noqa: WPS220
                    f"An invalid parameter {key!r} was specified in the 'where' "
                    "field. You cannot use aggregations or 'groupBy' clauses in 'where'",
                )



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from onetl.base.base_connection import BaseConnection
from onetl.base.base_db_connection import BaseDBConnection, BaseDBDialect
from onetl.base.base_file_connection import BaseFileConnection
from onetl.base.base_file_df_connection import (
    BaseFileDFConnection,
    FileDFReadOptions,
    FileDFWriteOptions,
)
from onetl.base.base_file_filter import BaseFileFilter
from onetl.base.base_file_format import BaseReadableFileFormat, BaseWritableFileFormat
from onetl.base.base_file_limit import BaseFileLimit
from onetl.base.contains_exception import ContainsException
from onetl.base.contains_get_df_schema import ContainsGetDFSchemaMethod
from onetl.base.contains_get_min_max_values import ContainsGetMinMaxValues
from onetl.base.path_protocol import PathProtocol, PathWithStatsProtocol
from onetl.base.path_stat_protocol import PathStatProtocol
from onetl.base.pure_path_protocol import PurePathProtocol
from onetl.base.supports_rename_dir import SupportsRenameDir



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import os

from humanize import naturalsize

from onetl.impl import BaseModel

# Metrics themselves are considered a part of driver result,
# ignore if result is smaller than 1MB
MIN_DRIVER_BYTES = 1_000_000


class SparkDriverMetrics(BaseModel):
    in_memory_bytes: int = 0

    @property
    def is_empty(self) -> bool:
        return self.in_memory_bytes < MIN_DRIVER_BYTES

    def update(self, other: SparkDriverMetrics) -> SparkDriverMetrics:
        self.in_memory_bytes += other.in_memory_bytes
        return self

    @property
    def details(self) -> str:
        if self.is_empty:
            return "No data"

        result = []
        if self.in_memory_bytes >= MIN_DRIVER_BYTES:
            result.append(f"In-memory data (approximate): {naturalsize(self.in_memory_bytes)}")

        return os.linesep.join(result)

    def __str__(self):
        return self.details



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import logging

from onetl.base import BaseFileLimit, PathProtocol
from onetl.impl import FrozenModel

log = logging.getLogger(__name__)


class MaxFilesCount(BaseFileLimit, FrozenModel):
    """Limits the total number of files handled by :ref:`file-downloader` or :ref:`file-mover`.

    .. versionadded:: 0.8.0
        Replaces deprecated ``onetl.core.FileLimit``

    Parameters
    ----------

    limit : int

        All files until ``limit`` (including) will be downloaded/moved, but ``limit+1`` will not.

    Examples
    --------

    Create filter which allows to handle 100 files, but stops on 101:

    .. code:: python

        from onetl.file.limit import MaxFilesCount

        limit = MaxFilesCount(100)
    """

    limit: int

    _counter: int = 0

    def __init__(self, limit: int):
        # this is only to allow passing glob as positional argument
        super().__init__(limit=limit)  # type: ignore

    def __repr__(self):
        return f"{self.__class__.__name__}({self.limit})"

    def reset(self):
        self._counter = 0
        return self

    def stops_at(self, path: PathProtocol) -> bool:
        if self.is_reached:
            return True

        if path.is_dir():
            # directories count does not matter
            return False

        self._counter += 1
        return self.is_reached

    @property
    def is_reached(self) -> bool:
        return self._counter >= self.limit



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import warnings
from typing import ClassVar, Optional

from etl_entities.instance import Host

from onetl._util.classproperty import classproperty
from onetl._util.version import Version
from onetl.connection.db_connection.jdbc_connection import JDBCConnection
from onetl.connection.db_connection.mysql.dialect import MySQLDialect
from onetl.connection.db_connection.mysql.options import (
    MySQLExecuteOptions,
    MySQLFetchOptions,
    MySQLReadOptions,
    MySQLSQLOptions,
    MySQLWriteOptions,
)
from onetl.hooks import slot, support_hooks
from onetl.impl.generic_options import GenericOptions

# do not import PySpark here, as we allow user to use `MySQL.get_packages()` for creating Spark session


class MySQLExtra(GenericOptions):
    useUnicode: str = "yes"  # noqa: N815
    characterEncoding: str = "UTF-8"  # noqa: N815

    class Config:
        extra = "allow"


@support_hooks
class MySQL(JDBCConnection):
    """MySQL JDBC connection. |support_hooks|

    Based on Maven package `com.mysql:mysql-connector-j:9.0.0 <https://mvnrepository.com/artifact/com.mysql/mysql-connector-j/9.0.0>`_
    (`official MySQL JDBC driver <https://dev.mysql.com/downloads/connector/j/8.4.html>`_).

    .. seealso::

        Before using this connector please take into account :ref:`mysql-prerequisites`

    .. versionadded:: 0.1.0

    Parameters
    ----------
    host : str
        Host of MySQL database. For example: ``mysql0012.domain.com`` or ``192.168.1.11``

    port : int, default: ``3306``
        Port of MySQL database

    user : str
        User, which have proper access to the database. For example: ``some_user``

    password : str
        Password for database connection

    database : str
        Database in RDBMS, NOT schema.

        See `this page <https://www.educba.com/postgresql-database-vs-schema/>`_ for more details

    spark : :obj:`pyspark.sql.SparkSession`
        Spark session.

    extra : dict, default: ``None``
        Specifies one or more extra parameters by which clients can connect to the instance.

        For example: ``{"useSSL": "false", "allowPublicKeyRetrieval": "true"}``

        See `MySQL JDBC driver properties documentation
        <https://dev.mysql.com/doc/connector-j/en/connector-j-reference-configuration-properties.html>`_
        for more details

    Examples
    --------

    MySQL connection initialization

    .. code:: python

        from onetl.connection import MySQL
        from pyspark.sql import SparkSession

        # Create Spark session with MySQL driver loaded
        maven_packages = MySQL.get_packages()
        spark = (
            SparkSession.builder.appName("spark-app-name")
            .config("spark.jars.packages", ",".join(maven_packages))
            .getOrCreate()
        )

        # Create connection
        mysql = MySQL(
            host="database.host.or.ip",
            user="user",
            password="*****",
            extra={"useSSL": "false", "allowPublicKeyRetrieval": "true"},
            spark=spark,
        )

    """

    host: Host
    port: int = 3306
    database: Optional[str] = None
    extra: MySQLExtra = MySQLExtra()

    ReadOptions = MySQLReadOptions
    WriteOptions = MySQLWriteOptions
    SQLOptions = MySQLSQLOptions
    FetchOptions = MySQLFetchOptions
    ExecuteOptions = MySQLExecuteOptions

    Extra = MySQLExtra
    Dialect = MySQLDialect

    DRIVER: ClassVar[str] = "com.mysql.cj.jdbc.Driver"

    @slot
    @classmethod
    def get_packages(cls, package_version: str | None = None) -> list[str]:
        """
        Get package names to be downloaded by Spark. Allows specifying a custom JDBC driver version for MySQL. |support_hooks|

        .. versionadded:: 0.9.0

        Parameters
        ----------
        package_version : str, optional
            Specifies the version of the MySQL JDBC driver to use. Defaults to ``9.0.0``.

            .. versionadded:: 0.11.0

        Examples
        --------
        .. code:: python

            from onetl.connection import MySQL

            MySQL.get_packages()

            # specify a custom package version
            MySQL.get_packages(package_version="8.2.0")
        """
        default_version = "9.0.0"
        version = Version(package_version or default_version).min_digits(3)

        return [f"com.mysql:mysql-connector-j:{version}"]

    @classproperty
    def package(cls) -> str:
        """Get package name to be downloaded by Spark."""
        msg = "`MySQL.package` will be removed in 1.0.0, use `MySQL.get_packages()` instead"
        warnings.warn(msg, UserWarning, stacklevel=3)
        return "com.mysql:mysql-connector-j:9.0.0"

    @property
    def jdbc_url(self) -> str:
        if self.database:
            return f"jdbc:mysql://{self.host}:{self.port}/{self.database}"

        return f"jdbc:mysql://{self.host}:{self.port}"

    @property
    def jdbc_params(self) -> dict:
        result = super().jdbc_params
        result.update(self.extra.dict(by_alias=True))
        return result

    @property
    def instance_url(self) -> str:
        return f"{self.__class__.__name__.lower()}://{self.host}:{self.port}"

    def __str__(self):
        return f"{self.__class__.__name__}[{self.host}:{self.port}]"



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import warnings


class SupportColumns:
    def validate_columns(
        self,
        columns: str | list[str] | None,
    ) -> list[str] | None:
        if columns is None:
            return ["*"]

        if isinstance(columns, str):
            warnings.warn(
                f"Passing DBReader(columns={columns!r}) is deprecated since v0.10.0 and will be removed in v1.0.0. "
                f"Use DBReader(columns=[{columns!r}] instead",
                category=UserWarning,
                stacklevel=3,
            )
            return [columns]

        return list(columns)



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import warnings
from typing import ClassVar, Optional

from etl_entities.instance import Host

from onetl._util.classproperty import classproperty
from onetl._util.spark import stringify
from onetl._util.version import Version
from onetl.connection.db_connection.jdbc_connection import JDBCConnection
from onetl.connection.db_connection.teradata.dialect import TeradataDialect
from onetl.connection.db_connection.teradata.options import (
    TeradataExecuteOptions,
    TeradataFetchOptions,
    TeradataReadOptions,
    TeradataSQLOptions,
    TeradataWriteOptions,
)
from onetl.hooks import slot
from onetl.impl import GenericOptions

# do not import PySpark here, as we allow user to use `Teradata.get_packages()` for creating Spark session


class TeradataExtra(GenericOptions):
    CHARSET: str = "UTF8"
    COLUMN_NAME: str = "ON"
    FLATTEN: str = "ON"
    MAYBENULL: str = "ON"
    STRICT_NAMES: str = "OFF"

    class Config:
        extra = "allow"
        prohibited_options = frozenset(("DATABASE", "DBS_PORT"))


class Teradata(JDBCConnection):
    """Teradata JDBC connection. |support_hooks|

    Based on package `com.teradata.jdbc:terajdbc:17.20.00.15 <https://central.sonatype.com/artifact/com.teradata.jdbc/terajdbc/17.20.00.15>`_
    (`official Teradata JDBC driver <https://downloads.teradata.com/download/connectivity/jdbc-driver>`_).

    .. seealso::

        Before using this connector please take into account :ref:`teradata-prerequisites`

    .. versionadded:: 0.1.0

    Parameters
    ----------
    host : str
        Host of Teradata database. For example: ``test.teradata.domain.com`` or ``193.168.1.12``

    port : int, default: ``1025``
        Port of Teradata database

    user : str
        User, which have proper access to the database. For example: ``some_user``

    password : str
        Password for database connection

    database : str, optional
        Database in RDBMS, NOT schema.

        See `this page <https://www.educba.com/postgresql-database-vs-schema/>`_ for more details

    spark : :obj:`pyspark.sql.SparkSession`
        Spark session.

    extra : dict, default: ``None``
        Specifies one or more extra parameters which should be appended to a connection string.

        For example: ``{"TMODE": "TERA", "MAYBENULL": "ON", "CHARSET": "UTF8", "LOGMECH":"LDAP"}``

        See `Teradata JDBC driver documentation
        <https://teradata-docs.s3.amazonaws.com/doc/connectivity/jdbc/reference/current/jdbcug_chapter_2.html#BABJIHBJ>`_
        for more details

        .. note::

            By default, these options are added to extra:

                * ``CHARSET = "UTF8"``
                * ``COLUMN_NAME = "ON"`` - allow reading column title from a table
                * ``FLATTEN = "ON"`` - improves error messages
                * ``MAYBENULL = "ON"``
                * ``STRICT_NAMES = "OFF"`` - ignore Spark options passed to JDBC URL

            It is possible to override default values, for example set ``extra={"FLATTEN": "OFF"}``

    Examples
    --------

    Teradata connection with LDAP auth:

    .. code:: python

        from onetl.connection import Teradata
        from pyspark.sql import SparkSession

        # Create Spark session with Teradata driver loaded
        maven_packages = Teradata.get_packages()
        spark = (
            SparkSession.builder.appName("spark-app-name")
            .config("spark.jars.packages", ",".join(maven_packages))
            .getOrCreate()
        )

        # Create connection
        teradata = Teradata(
            host="database.host.or.ip",
            user="user",
            password="*****",
            extra={
                "TMODE": "TERA",  # "TERA" or "ANSI"
                "LOGMECH": "LDAP",
                "LOG": "TIMING",  # increase log level
            },
            spark=spark,
        )

    """

    host: Host
    port: int = 1025
    database: Optional[str] = None
    extra: TeradataExtra = TeradataExtra()

    ReadOptions = TeradataReadOptions
    WriteOptions = TeradataWriteOptions
    SQLOptions = TeradataSQLOptions
    FetchOptions = TeradataFetchOptions
    ExecuteOptions = TeradataExecuteOptions

    Extra = TeradataExtra
    Dialect = TeradataDialect

    DRIVER: ClassVar[str] = "com.teradata.jdbc.TeraDriver"
    _CHECK_QUERY: ClassVar[str] = "SELECT 1 AS check_result"

    @slot
    @classmethod
    def get_packages(
        cls,
        package_version: str | None = None,
    ) -> list[str]:
        """
        Get package names to be downloaded by Spark. Allows specifying custom JDBC driver versions for Teradata. |support_hooks|

        .. versionadded:: 0.9.0

        Parameters
        ----------
        package_version : str, optional
            Specifies the version of the Teradata JDBC driver to use. Defaults to ``17.20.00.15``.

            .. versionadded:: 0.11.0

        Examples
        --------
        .. code:: python

            from onetl.connection import Teradata

            Teradata.get_packages()

            # specify custom driver version
            Teradata.get_packages(package_version="20.00.00.18")
        """
        default_package_version = "17.20.00.15"
        version = Version(package_version or default_package_version).min_digits(4)

        return [f"com.teradata.jdbc:terajdbc:{version}"]

    @classproperty
    def package(cls) -> str:
        """Get package name to be downloaded by Spark."""
        msg = "`Teradata.package` will be removed in 1.0.0, use `Teradata.get_packages()` instead"
        warnings.warn(msg, UserWarning, stacklevel=3)
        return "com.teradata.jdbc:terajdbc:17.20.00.15"

    @property
    def jdbc_url(self) -> str:
        # Teradata JDBC driver documentation specifically mentions that params from
        # java.sql.DriverManager.getConnection(url, params) are used to only retrieve 'user' and 'password' values.
        # Other params should be passed via url
        properties = self.extra.dict(by_alias=True)

        if self.database:
            properties["DATABASE"] = self.database

        properties["DBS_PORT"] = self.port

        connection_params = []
        for key, value in sorted(properties.items()):
            string_value = stringify(value)
            if "," in string_value:
                connection_params.append(f"{key}='{string_value}'")
            else:
                connection_params.append(f"{key}={string_value}")

        return f"jdbc:teradata://{self.host}/{','.join(connection_params)}"

    @property
    def instance_url(self) -> str:
        return f"{self.__class__.__name__.lower()}://{self.host}:{self.port}"

    def __str__(self):
        return f"{self.__class__.__name__}[{self.host}:{self.port}]"



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum


class SparkListenerTaskStatus(str, Enum):
    PENDING = "PENDING"
    RUNNING = "RUNNING"
    SUCCESS = "SUCCESS"
    FAILED = "FAILED"
    KILLED = "KILLED"

    def __str__(self):
        return self.value


@dataclass
class SparkListenerTaskInputMetrics:
    bytes_read: int = 0
    records_read: int = 0

    @classmethod
    def create(cls, task_input_metrics):
        return cls(
            bytes_read=task_input_metrics.bytesRead(),
            records_read=task_input_metrics.recordsRead(),
        )


@dataclass
class SparkListenerTaskOutputMetrics:
    bytes_written: int = 0
    records_written: int = 0

    @classmethod
    def create(cls, task_output_metrics):
        return cls(
            bytes_written=task_output_metrics.bytesWritten(),
            records_written=task_output_metrics.recordsWritten(),
        )


@dataclass
class SparkListenerTaskMetrics:
    """Python representation of Spark TaskMetrics object.

    See `documentation <https://spark.apache.org/docs/latest/monitoring.html#executor-task-metrics>`_.
    """

    executor_run_time_milliseconds: int = 0
    executor_cpu_time_nanoseconds: int = 0
    peak_execution_memory_bytes: int = 0
    memory_spilled_bytes: int = 0
    disk_spilled_bytes: int = 0
    result_size_bytes: int = 0
    input_metrics: SparkListenerTaskInputMetrics = field(default_factory=SparkListenerTaskInputMetrics)
    output_metrics: SparkListenerTaskOutputMetrics = field(default_factory=SparkListenerTaskOutputMetrics)

    @classmethod
    def create(cls, task_metrics):
        return cls(
            executor_run_time_milliseconds=task_metrics.executorRunTime(),
            executor_cpu_time_nanoseconds=task_metrics.executorCpuTime(),
            peak_execution_memory_bytes=task_metrics.peakExecutionMemory(),
            memory_spilled_bytes=task_metrics.memoryBytesSpilled(),
            disk_spilled_bytes=task_metrics.diskBytesSpilled(),
            result_size_bytes=task_metrics.resultSize(),
            input_metrics=SparkListenerTaskInputMetrics.create(task_metrics.inputMetrics()),
            output_metrics=SparkListenerTaskOutputMetrics.create(task_metrics.outputMetrics()),
        )


@dataclass
class SparkListenerTask:
    id: int
    status: SparkListenerTaskStatus = SparkListenerTaskStatus.PENDING
    metrics: SparkListenerTaskMetrics | None = field(default=None, repr=False, init=False)

    @classmethod
    def create(cls, task_info):
        # https://spark.apache.org/docs/3.5.2/api/java/org/apache/spark/scheduler/TaskInfo.html
        return cls(id=task_info.taskId())

    def on_task_start(self, event):
        # https://spark.apache.org/docs/3.5.2/api/java/org/apache/spark/scheduler/SparkListenerTaskStart.html
        self.status = SparkListenerTaskStatus(event.taskInfo().status())

    def on_task_end(self, event):
        # https://spark.apache.org/docs/3.5.2/api/java/org/apache/spark/scheduler/SparkListenerTaskEnd.html
        self.status = SparkListenerTaskStatus(event.taskInfo().status())
        self.metrics = SparkListenerTaskMetrics.create(event.taskMetrics())



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from onetl.connection.db_connection.jdbc_mixin.connection import (
    JDBCMixin,
    JDBCStatementType,
)
from onetl.connection.db_connection.jdbc_mixin.options import JDBCOptions



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass, field
from enum import Enum

from onetl._metrics.listener.job import SparkListenerJob, SparkListenerJobStatus


class SparkListenerExecutionStatus(str, Enum):
    STARTED = "STARTED"
    COMPLETE = "COMPLETE"
    FAILED = "FAILED"

    def __str__(self):
        return self.value


class SparkSQLMetricNames(str, Enum):  # noqa: WPS338
    # Metric names passed to SQLMetrics.createMetric(...)
    # But only those we're interested in.

    # https://github.com/apache/spark/blob/v3.5.2/sql/core/src/main/scala/org/apache/spark/sql/execution/DataSourceScanExec.scala#L231
    NUMBER_OF_PARTITIONS_READ = "number of partitions read"

    # https://github.com/apache/spark/blob/v3.5.2/sql/core/src/main/scala/org/apache/spark/sql/execution/DataSourceScanExec.scala#L225-L227
    NUMBER_OF_FILES_READ = "number of files read"
    SIZE_OF_FILES_READ = "size of files read"

    # https://github.com/apache/spark/blob/v3.5.2/sql/core/src/main/scala/org/apache/spark/sql/execution/DataSourceScanExec.scala#L225-L227
    STATIC_NUMBER_OF_FILES_READ = "static number of files read"
    STATIC_SIZE_OF_FILES_READ = "static size of files read"

    # https://github.com/apache/spark/blob/v3.5.2/sql/core/src/main/scala/org/apache/spark/sql/execution/datasources/BasicWriteStatsTracker.scala#L241-L246
    NUMBER_OF_DYNAMIC_PART = "number of dynamic part"
    NUMBER_OF_WRITTEN_FILES = "number of written files"

    def __str__(self):
        return self.value

    @classmethod
    def values(cls):
        return set(cls.__members__.values())


@dataclass
class SparkListenerExecution:
    id: int
    description: str | None = None
    external_id: str | None = None
    status: SparkListenerExecutionStatus = SparkListenerExecutionStatus.STARTED

    # These metrics are emitted by any command performed within this execution, so we can have multiple values.
    # Some metrics can be summarized, but some not, so we store a list.
    metrics: dict[SparkSQLMetricNames, list[str]] = field(default_factory=lambda: defaultdict(list), repr=False)

    _jobs: dict[int, SparkListenerJob] = field(default_factory=dict, repr=False, init=False)

    @property
    def jobs(self) -> list[SparkListenerJob]:
        result = []
        for job_id in sorted(self._jobs.keys()):
            result.append(self._jobs[job_id])
        return result

    def on_execution_start(self, event):
        # https://github.com/apache/spark/blob/v3.5.2/sql/core/src/main/scala/org/apache/spark/sql/execution/ui/SQLListener.scala#L44-L58
        self.status = SparkListenerExecutionStatus.STARTED

    def on_execution_end(self, event):
        # https://github.com/apache/spark/blob/v3.5.2/sql/core/src/main/scala/org/apache/spark/sql/execution/ui/SQLListener.scala#L61-L83
        for job in self._jobs.values():
            if job.status == SparkListenerJobStatus.FAILED:
                self.status = SparkListenerExecutionStatus.FAILED
                break
        else:
            self.status = SparkListenerExecutionStatus.COMPLETE

    def on_job_start(self, event):
        job_id = event.jobId()
        job = SparkListenerJob.create(event)
        self._jobs[job_id] = job
        job.on_job_start(event)

    def on_job_end(self, event):
        job_id = event.jobId()
        job = self._jobs.get(job_id)

        if job:
            job.on_job_end(event)

        # in some cases Execution consists of just one job with same id
        if job_id == self.id:
            self.on_execution_end(event)

    # push down events
    def on_stage_start(self, event):
        for job in self._jobs.values():
            job.on_stage_start(event)

    def on_stage_end(self, event):
        for job in self._jobs.values():
            job.on_stage_end(event)

    def on_task_start(self, event):
        for job in self._jobs.values():
            job.on_task_start(event)

    def on_task_end(self, event):
        for job in self._jobs.values():
            job.on_task_end(event)



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import os
from dataclasses import astuple, dataclass
from typing import Generic, Sequence, TypeVar

from onetl.base import PurePathProtocol

PurePath = TypeVar("PurePath", bound=PurePathProtocol)


@dataclass(eq=False, frozen=True)
class PathContainer(Generic[PurePath]):
    """
    Read-only container for ``pathlib.PurePath``-like classes with some custom logic.

    This is need to avoid creating objects of current class using ``PurePath``-like methods
    (e.g. ``path.with_name("new_name")``) or operators (e.g. ``path / "sub_path"``).

    It proxies all the method calls and operators to the underlying ``path`` field,
    so every method or operator returns new object of the same class as ``path`` field, without any magic.
    """

    path: PurePath

    def __str__(self) -> str:
        return str(self.path)

    def __bytes__(self):
        return os.fsencode(self.path)

    def __fspath__(self) -> str:
        return os.fspath(self.path)

    def __truediv__(self, other) -> PurePath:
        return self.path / other

    def __rtruediv__(self, other) -> PurePath:
        return other / self.path

    def __hash__(self):
        return hash(self.path)

    def __eq__(self, other):
        if not isinstance(other, PathContainer):
            return self.path == other

        return astuple(self, tuple_factory=self._compare_tuple) == astuple(other, tuple_factory=self._compare_tuple)

    def __lt__(self, other):
        if not isinstance(other, PathContainer):
            return self.path < other

        return self.path < other.path

    def __le__(self, other):
        if not isinstance(other, PathContainer):
            return self.path <= other

        return self.path <= other.path

    def __gt__(self, other):
        if not isinstance(other, PathContainer):
            return self.path > other

        return self.path > other.path

    def __ge__(self, other):
        if not isinstance(other, PathContainer):
            return self.path >= other

        return self.path >= other.path

    # typing_extensions 4.6+ implementation of `isinstance`` does not use __getattr__ anymore,
    # so we need to create methods explicitly
    @property
    def name(self) -> str:
        return self.path.name

    @property
    def parent(self) -> PurePathProtocol | PurePath:
        return self.path.parent

    @property
    def parents(self) -> Sequence[PurePathProtocol | PurePath]:
        return self.path.parents

    @property
    def parts(self) -> Sequence[str]:
        return self.path.parts

    def is_absolute(self) -> bool:
        return self.path.is_absolute()

    def match(self, path_pattern) -> bool:
        return self.path.match(path_pattern)

    def relative_to(self, *other) -> PurePathProtocol | PurePath:
        return self.path.relative_to(*other)

    def as_posix(self) -> str:
        return self.path.as_posix()

    def joinpath(self, *args) -> PurePathProtocol | PurePath:
        return self.path.joinpath(*args)

    def _compare_tuple(self, args) -> tuple:
        return tuple(args)



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import logging
from typing import Iterable

from onetl.base import BaseFileLimit, PathProtocol

log = logging.getLogger(__name__)


def limits_stop_at(path: PathProtocol, limits: Iterable[BaseFileLimit]) -> bool:
    """
    Check if some of limits stops at given path.

    .. versionadded:: 0.8.0

    Parameters
    ----------
    path : :obj:`onetl.base.path_protocol.PathProtocol`
        Path to check.

    limits : Iterable of :obj:`onetl.base.base_file_limit.BaseFileLimit`
        Limits to test path against.

    Returns
    -------
    ``True`` if any of limit is reached while handling the path, ``False`` otherwise.

    If no limits are passed, returns ``False``.

    Examples
    --------

    >>> from onetl.file.limit import MaxFilesCount, limits_stop_at
    >>> from onetl.impl import LocalPath
    >>> limits = [MaxFilesCount(2)]
    >>> limits_stop_at(LocalPath("/path/to/file.csv"), limits)
    False
    >>> limits_stop_at(LocalPath("/path/to/file.csv"), limits)
    True
    """
    reached = []
    for limit in limits:
        if limit.stops_at(path):
            reached.append(limit)

    if reached:
        log.debug("|FileLimit| Limits %r are reached", reached)
        return True

    return False



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from typing import Sequence, TypeVar

from typing_extensions import Protocol, runtime_checkable

T = TypeVar("T", bound="PurePathProtocol", covariant=True)


@runtime_checkable
class PurePathProtocol(Protocol[T]):
    """
    Generic protocol for :obj:`pathlib.PurePath` like objects.

    Includes only minimal set of methods which allow to get path items, like parent, name, etc
    """

    def __fspath__(self) -> str:
        """
        Get string representation of path
        """

    def __eq__(self, other) -> bool:
        """
        Check if two paths are equal
        """

    def __hash__(self) -> int:
        """
        Get hash value for path
        """

    def __truediv__(self, key) -> T:
        """
        Add items to path
        """

    def __rtruediv__(self, key) -> T:
        """
        Add items to path
        """

    @property
    def name(self) -> str:
        """
        Get path name
        """

    @property
    def parent(self) -> T:
        """
        Get parent path
        """

    @property
    def parents(self) -> Sequence[T]:
        """
        Get parent paths
        """

    @property
    def parts(self) -> Sequence[str]:
        """
        Get path parts
        """

    def is_absolute(self) -> bool:
        """
        Checks if this path is absolute
        """

    def match(self, path_pattern) -> bool:
        """
        Checks if path matches a glob pattern
        """

    def relative_to(self, *other) -> T:
        """
        Return the relative path to another path
        """

    def as_posix(self) -> str:
        """
        Get POSIX representation of path
        """

    def joinpath(self, *args) -> T:
        """
        Add items to path
        """



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from typing import TYPE_CHECKING

from onetl._metrics.command import SparkCommandMetrics
from onetl._metrics.extract import extract_metrics_from_execution
from onetl._metrics.listener import SparkMetricsListener

if TYPE_CHECKING:
    from pyspark.sql import SparkSession


class SparkMetricsRecorder:
    def __init__(self, spark: SparkSession):
        self._listener = SparkMetricsListener(spark=spark)

    def __enter__(self):
        self._listener.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._listener.__exit__(exc_type, exc_val, exc_tb)

    def metrics(self) -> SparkCommandMetrics:
        result = SparkCommandMetrics()
        for execution in self._listener.executions:
            result = result.update(extract_metrics_from_execution(execution))
        return result



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from logging import getLogger
from typing import TYPE_CHECKING

try:
    from pydantic.v1 import Field, validator
except (ImportError, AttributeError):
    from pydantic import Field, validator  # type: ignore[no-redef, assignment]

from onetl._util.spark import try_import_pyspark
from onetl.base import BaseDBConnection
from onetl.connection.db_connection.db_connection.dialect import DBDialect
from onetl.impl import FrozenModel
from onetl.log import log_with_indent

if TYPE_CHECKING:
    from pyspark.sql import SparkSession

log = getLogger(__name__)


class DBConnection(BaseDBConnection, FrozenModel):
    spark: SparkSession = Field(repr=False)

    Dialect = DBDialect

    @classmethod
    def _forward_refs(cls) -> dict[str, type]:
        try_import_pyspark()

        from pyspark.sql import SparkSession  # noqa: WPS442

        # avoid importing pyspark unless user called the constructor,
        # as we allow user to use `Connection.get_packages()` for creating Spark session
        refs = super()._forward_refs()
        refs["SparkSession"] = SparkSession
        return refs

    @validator("spark")
    def _check_spark_session_alive(cls, spark):
        # https://stackoverflow.com/a/36044685
        msg = "Spark session is stopped. Please recreate Spark session."
        try:
            if not spark._jsc.sc().isStopped():
                return spark
        except Exception as e:
            # None has no attribute "something"
            raise ValueError(msg) from e

        raise ValueError(msg)

    def _log_parameters(self):
        log.info("|%s| Using connection parameters:", self.__class__.__name__)
        parameters = self.dict(exclude_none=True, exclude={"spark"})
        for attr, value in parameters.items():
            log_with_indent(log, "%s = %r", attr, value)



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import json
import logging
from contextlib import closing
from typing import TYPE_CHECKING, Any, List, Optional

from etl_entities.instance import Cluster

try:
    from pydantic.v1 import root_validator, validator
except (ImportError, AttributeError):
    from pydantic import root_validator, validator  # type: ignore[no-redef, assignment]

from onetl._util.java import try_import_java_class
from onetl._util.scala import get_default_scala_version
from onetl._util.spark import get_spark_version, stringify
from onetl._util.version import Version
from onetl.connection.db_connection.db_connection import DBConnection
from onetl.connection.db_connection.kafka.dialect import KafkaDialect
from onetl.connection.db_connection.kafka.extra import KafkaExtra
from onetl.connection.db_connection.kafka.kafka_auth import KafkaAuth
from onetl.connection.db_connection.kafka.kafka_basic_auth import KafkaBasicAuth
from onetl.connection.db_connection.kafka.kafka_kerberos_auth import KafkaKerberosAuth
from onetl.connection.db_connection.kafka.kafka_plaintext_protocol import (
    KafkaPlaintextProtocol,
)
from onetl.connection.db_connection.kafka.kafka_protocol import KafkaProtocol
from onetl.connection.db_connection.kafka.kafka_scram_auth import KafkaScramAuth
from onetl.connection.db_connection.kafka.kafka_ssl_protocol import KafkaSSLProtocol
from onetl.connection.db_connection.kafka.options import (
    KafkaReadOptions,
    KafkaTopicExistBehaviorKafka,
    KafkaWriteOptions,
)
from onetl.connection.db_connection.kafka.slots import KafkaSlots
from onetl.exception import MISSING_JVM_CLASS_MSG, TargetAlreadyExistsError
from onetl.hooks import slot, support_hooks
from onetl.hwm.window import Window
from onetl.log import log_collection, log_with_indent

if TYPE_CHECKING:
    from pyspark.sql import DataFrame
    from pyspark.sql.types import StructType

log = logging.getLogger(__name__)


@support_hooks
class Kafka(DBConnection):
    """
    This connector is designed to read and write from Kafka in batch mode.

    Based on `official Kafka Source For Spark
    <https://spark.apache.org/docs/latest/structured-streaming-kafka-integration.html>`_.

    .. seealso::

        Before using this connector please take into account :ref:`kafka-prerequisites`

    .. note::

        This connector is for **batch** ETL processes, not streaming.

    .. versionadded:: 0.9.0

    Parameters
    ----------

    addresses : list[str]
        A list of broker addresses, for example ``["192.168.1.10:9092", "192.168.1.11:9092"]``.

    cluster : str
        Cluster name. Used for HWM and lineage.

    auth : KafkaAuth, default: ``None``
        Kafka authentication mechanism. ``None`` means anonymous auth.

    protocol : KafkaProtocol, default: :obj:`PlaintextProtocol <onetl.connection.db_connection.kafka.kafka_plaintext_protocol.KafkaPlaintextProtocol>`
        Kafka security protocol.

    extra : dict, default: ``None``
        A dictionary of additional properties to be used when connecting to Kafka.

        These are Kafka-specific properties that control behavior of the producer or consumer. See:

        * `producer options documentation <https://kafka.apache.org/documentation/#producerconfigs>`_
        * `consumer options documentation <https://kafka.apache.org/documentation/#consumerconfigs>`_
        * `Spark Kafka documentation <https://spark.apache.org/docs/latest/structured-streaming-kafka-integration.html#kafka-specific-configurations>`_

        Options are passed without ``kafka.`` prefix, for example:

        For example:

        .. code:: python

            extra = {
                "group.id": "myGroup",
                "request.timeout.ms": 120000,
            }

        .. warning::

            Options that populated from connection
            attributes (like ``bootstrap.servers``, ``sasl.*``, ``ssl.*``) are not allowed to override.

        .. note::

            At current version Kafka connection doesn't support batch strategies.

    Examples
    --------

    Connect to Kafka using ``PLAINTEXT`` protocol and without any auth mechanism (anonymous user, default):

    .. code:: python

        from onetl.connection import Kafka
        from pyspark.sql import SparkSession

        # Create Spark session with Kafka connector loaded
        maven_packages = Kafka.get_packages(spark_version="3.2.4")
        spark = (
            SparkSession.builder.appName("spark-app-name")
            .config("spark.jars.packages", ",".join(maven_packages))
            .getOrCreate()
        )

        # Create connection
        kafka = Kafka(
            addresses=["mybroker:9092", "anotherbroker:9092"],
            cluster="my-cluster",
            spark=spark,
        )

    Connect to Kafka using ``PLAINTEXT`` protocol and basic (``PLAIN``) auth mechanism:

    .. code:: python

        # Create Spark session with Kafka connector loaded
        ...

        # Create connection
        kafka = Kafka(
            addresses=["mybroker:9092", "anotherbroker:9092"],
            cluster="my-cluster",
            auth=Kafka.BasicAuth(
                user="me",
                password="password",
            ),
            spark=spark,
        )

    Connect to Kafka using ``PLAINTEXT`` protocol and Kerberos (``GSSAPI``) auth mechanism:

    .. code:: python

        # Create Spark session with Kafka connector loaded
        ...

        # Create connection
        kafka = Kafka(
            addresses=["mybroker:9092", "anotherbroker:9092"],
            cluster="my-cluster",
            auth=Kafka.KerberosAuth(
                principal="me@example.com",
                keytab="/path/to/keytab",
                deploy_keytab=True,
            ),
            spark=spark,
        )

    Connect to Kafka using ``SASL_SSL`` protocol and ``SCRAM-SHA-512`` auth mechanism:

    .. code:: python

        from pathlib import Path

        # Create Spark session with Kafka connector loaded
        ...

        # Create connection
        kafka = Kafka(
            addresses=["mybroker:9092", "anotherbroker:9092"],
            cluster="my-cluster",
            protocol=Kafka.SSLProtocol(
                keystore_type="PEM",
                keystore_certificate_chain=Path("path/to/user.crt").read_text(),
                keystore_key=Path("path/to/user.key").read_text(),
                truststore_type="PEM",
                truststore_certificates=Path("/path/to/server.crt").read_text(),
            ),
            auth=Kafka.ScramAuth(
                user="me",
                password="abc",
                digest="SHA-512",
            ),
            spark=spark,
        )

    Connect to Kafka with extra options:

    .. code:: python

        # Create Spark session with Kafka connector loaded
        ...

        # Create connection
        kafka = Kafka(
            addresses=["mybroker:9092", "anotherbroker:9092"],
            cluster="my-cluster",
            protocol=...,
            auth=...,
            extra={"max.request.size": 1000000},
            spark=spark,
        )

    """

    BasicAuth = KafkaBasicAuth
    KerberosAuth = KafkaKerberosAuth
    ScramAuth = KafkaScramAuth
    ReadOptions = KafkaReadOptions
    WriteOptions = KafkaWriteOptions
    SSLProtocol = KafkaSSLProtocol
    Extra = KafkaExtra
    Dialect = KafkaDialect
    PlaintextProtocol = KafkaPlaintextProtocol
    Slots = KafkaSlots

    cluster: Cluster
    addresses: List[str]
    auth: Optional[KafkaAuth] = None
    protocol: KafkaProtocol = PlaintextProtocol()
    extra: KafkaExtra = KafkaExtra()

    @slot
    def check(self):
        log.info("|%s| Checking connection availability...", self.__class__.__name__)
        self._log_parameters()

        try:
            self._get_topics()
            log.info("|%s| Connection is available.", self.__class__.__name__)
        except Exception as e:
            log.exception("|%s| Connection is unavailable", self.__class__.__name__)
            raise RuntimeError("Connection is unavailable") from e
        return self

    @slot
    def read_source_as_df(  # noqa:  WPS231
        self,
        source: str,
        columns: list[str] | None = None,
        hint: Any | None = None,
        where: Any | None = None,
        df_schema: StructType | None = None,
        window: Window | None = None,
        limit: int | None = None,
        options: KafkaReadOptions = KafkaReadOptions(),  # noqa: B008, WPS404
    ) -> DataFrame:
        log.info("|%s| Reading data from topic %r", self.__class__.__name__, source)
        if source not in self._get_topics():
            raise ValueError(f"Topic {source!r} doesn't exist")

        result_options = {f"kafka.{key}": value for key, value in self._get_connection_properties().items()}
        result_options.update(options.dict(by_alias=True, exclude_none=True))
        result_options["subscribe"] = source

        if window and window.expression == "offset":
            # the 'including' flag in window values are relevant for batch strategies which are not
            # supported by Kafka, therefore we always get offsets including border values
            starting_offsets = dict(window.start_from.value) if window.start_from.value else {}
            ending_offsets = dict(window.stop_at.value) if window.stop_at.value else {}

            # when the Kafka topic's number of partitions has increased during incremental processing,
            # new partitions, which are present in ending_offsets but not in
            # starting_offsets, are assigned a default offset (0 in this case).
            for partition in ending_offsets:
                if partition not in starting_offsets:
                    starting_offsets[partition] = 0

            if starting_offsets:
                result_options["startingOffsets"] = json.dumps({source: starting_offsets})
            if ending_offsets:
                result_options["endingOffsets"] = json.dumps({source: ending_offsets})

        df = self.spark.read.format("kafka").options(**result_options).load()

        if limit is not None:
            df = df.limit(limit)

        log.info("|%s| Dataframe is successfully created.", self.__class__.__name__)
        return df

    @slot
    def write_df_to_target(
        self,
        df: DataFrame,
        target: str,
        options: KafkaWriteOptions = KafkaWriteOptions(),  # noqa: B008, WPS404
    ) -> None:
        # Check that the DataFrame doesn't contain any columns not in the schema
        required_columns = {"value"}
        optional_columns = {"key", "partition", "headers"}
        allowed_columns = required_columns | optional_columns | {"topic"}
        df_columns = set(df.columns)
        if not df_columns.issubset(allowed_columns):
            invalid_columns = df_columns - allowed_columns
            raise ValueError(
                f"Invalid column names: {sorted(invalid_columns)}. "
                f"Expected columns: {sorted(required_columns)} (required),"
                f" {sorted(optional_columns)} (optional)",
            )

        # Check that the DataFrame doesn't contain a 'headers' column with includeHeaders=False
        if not options.include_headers and "headers" in df.columns:
            raise ValueError("Cannot write 'headers' column with kafka.WriteOptions(include_headers=False)")

        spark_version = get_spark_version(self.spark)
        if options.include_headers and spark_version.major < 3:
            raise ValueError(
                f"kafka.WriteOptions(include_headers=True) requires Spark 3.x, got {spark_version}",
            )

        if "topic" in df.columns:
            log.warning("The 'topic' column in the DataFrame will be overridden with value %r", target)

        write_options = {f"kafka.{key}": value for key, value in self._get_connection_properties().items()}
        write_options.update(options.dict(by_alias=True, exclude_none=True, exclude={"if_exists"}))
        write_options["topic"] = target

        # As of Apache Spark version 3.5.2, the mode 'error' is not functioning as expected.
        # This issue has been reported and can be tracked at:
        # https://issues.apache.org/jira/browse/SPARK-44774
        mode = options.if_exists
        if mode == KafkaTopicExistBehaviorKafka.ERROR and target in self._get_topics():
            raise TargetAlreadyExistsError(f"Topic {target} already exists")

        log.info("|%s| Saving data to a topic %r", self.__class__.__name__, target)
        df.write.format("kafka").mode(mode).options(**write_options).save()
        log.info("|%s| Data is successfully written to topic %r.", self.__class__.__name__, target)

    @slot
    def get_df_schema(
        self,
        source: str,
        columns: list[str] | None = None,
        options: KafkaReadOptions = KafkaReadOptions(),  # noqa:  WPS404
    ) -> StructType:
        from pyspark.sql.types import (  # noqa:  WPS442
            ArrayType,
            BinaryType,
            IntegerType,
            LongType,
            StringType,
            StructField,
            StructType,
            TimestampType,
        )

        all_fields = [
            StructField("key", BinaryType(), nullable=True),
            StructField("value", BinaryType(), nullable=False),
            StructField("topic", StringType(), nullable=True),
            StructField("partition", IntegerType(), nullable=True),
            StructField("offset", LongType(), nullable=True),
            StructField("timestamp", TimestampType(), nullable=True),
            StructField("timestampType", IntegerType(), nullable=True),
            StructField(
                "headers",
                ArrayType(
                    StructType(
                        [
                            StructField("key", StringType(), nullable=True),
                            StructField("value", BinaryType(), nullable=True),
                        ],
                    ),
                ),
                nullable=True,
            ),
        ]

        filtered_fields = [field for field in all_fields if columns is None or field.name in columns]

        return StructType(filtered_fields)

    @slot
    @classmethod
    def get_packages(
        cls,
        spark_version: str,
        scala_version: str | None = None,
    ) -> list[str]:
        """
        Get package names to be downloaded by Spark. |support_hooks|

        See `Maven package index <https://mvnrepository.com/artifact/org.apache.spark/spark-sql-kafka-0-10>`_
        for all available packages.

        Parameters
        ----------
        spark_version : str
            Spark version in format ``major.minor.patch``.

        scala_version : str, optional
            Scala version in format ``major.minor``.

            If ``None``, ``spark_version`` is used to determine Scala version.

        Examples
        --------

        .. code:: python

            from onetl.connection import Kafka

            Kafka.get_packages(spark_version="3.2.4")
            Kafka.get_packages(spark_version="3.2.4", scala_version="2.12")

        """

        # Connector version is same as Spark, do not perform any additional checks
        spark_ver = Version(spark_version).min_digits(3)
        if spark_ver < Version("2.4"):
            # Kafka connector for Spark 2.3 is build with Kafka client 0.10.0.1 which does not support
            # passing `sasl.jaas.config` option. It is supported only in 0.10.2.0,
            # see https://issues.apache.org/jira/browse/KAFKA-4259
            # Old client requires generating JAAS file and placing it to filesystem, which is not secure.
            raise ValueError(f"Spark version must be at least 2.4, got {spark_ver}")

        scala_ver = Version(scala_version).min_digits(2) if scala_version else get_default_scala_version(spark_ver)
        return [
            f"org.apache.spark:spark-sql-kafka-0-10_{scala_ver.format('{0}.{1}')}:{spark_ver.format('{0}.{1}.{2}')}",
        ]

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @slot
    def close(self):
        """
        Close all connections created to Kafka. |support_hooks|

        .. note::

            Connection can be used again after it was closed.

        Returns
        -------
        Connection itself

        Examples
        --------

        Close connection automatically:

        .. code:: python

            with connection:
                ...

        Close connection manually:

        .. code:: python

            connection.close()

        """
        self.protocol.cleanup(self)
        if self.auth:
            self.auth.cleanup(self)
        return self

    # Do not all __del__ with calling .close(), like other connections,
    # because this can influence dataframes created by this connection.
    # For example, .close() deletes local keytab copy.

    @slot
    def get_min_max_values(
        self,
        source: str,
        window: Window,
        hint: Any | None = None,
        where: Any | None = None,
        options: KafkaReadOptions | dict | None = None,
    ) -> tuple[dict[int, int], dict[int, int]]:
        log.info("|%s| Getting min and max offset values for topic %r ...", self.__class__.__name__, source)

        consumer = self._get_java_consumer()
        with closing(consumer):
            # https://kafka.apache.org/22/javadoc/org/apache/kafka/clients/consumer/KafkaConsumer.html#partitionsFor-java.lang.String-
            partition_infos = consumer.partitionsFor(source)

            jvm = self.spark._jvm  # type: ignore[attr-defined]
            topic_partitions = [
                jvm.org.apache.kafka.common.TopicPartition(source, p.partition())  # type: ignore[union-attr]
                for p in partition_infos
            ]

            # https://kafka.apache.org/22/javadoc/org/apache/kafka/clients/consumer/KafkaConsumer.html#beginningOffsets-java.util.Collection-
            beginning_offsets = consumer.beginningOffsets(topic_partitions)

            # https://kafka.apache.org/22/javadoc/org/apache/kafka/clients/consumer/KafkaConsumer.html#endOffsets-java.util.Collection-
            end_offsets = consumer.endOffsets(topic_partitions)

            min_offsets = {}
            max_offsets = {}
            for tp in topic_partitions:
                partition_id = tp.partition()
                begin_offset = beginning_offsets[tp]
                end_offset = end_offsets[tp]

                if window.start_from and window.start_from.is_set():
                    window_start = window.start_from.value.get(partition_id, begin_offset)
                    begin_offset = max(window_start, begin_offset)
                if window.stop_at and window.stop_at.is_set():
                    window_stop = window.stop_at.value.get(partition_id, end_offset)
                    end_offset = min(window_stop, end_offset)

                min_offsets[partition_id] = begin_offset
                max_offsets[partition_id] = end_offset

        log.info("|%s| Received min and max offset values for each partition.", self.__class__.__name__)
        for partition_id in sorted(min_offsets.keys()):
            log.debug(
                "|%s| Partition %d: Min Offset = %d, Max Offset = %d",
                self.__class__.__name__,
                partition_id,
                min_offsets[partition_id],
                max_offsets[partition_id],
            )

        return min_offsets, max_offsets

    @property
    def instance_url(self):
        return "kafka://" + self.cluster

    def __str__(self):
        return f"{self.__class__.__name__}[{self.cluster}]"

    @root_validator(pre=True)
    def _get_addresses_by_cluster(cls, values):
        cluster = values.get("cluster")
        addresses = values.get("addresses")
        if not addresses:
            cluster_addresses = cls.Slots.get_cluster_addresses(cluster) or []
            if cluster_addresses:
                log.debug("|%s| Set cluster %r addresses: %r", cls.__name__, cluster, cluster_addresses)
                values["addresses"] = cluster_addresses
            else:
                raise ValueError("Passed empty parameter 'addresses'")
        return values

    @validator("cluster")
    def _validate_cluster_name(cls, cluster):
        log.debug("|%s| Normalizing cluster %r name...", cls.__name__, cluster)
        validated_cluster = cls.Slots.normalize_cluster_name(cluster) or cluster
        if validated_cluster != cluster:
            log.debug("|%s|   Got %r", cls.__name__, validated_cluster)

        log.debug("|%s| Checking if cluster %r is a known cluster...", cls.__name__, validated_cluster)
        known_clusters = cls.Slots.get_known_clusters()
        if known_clusters and validated_cluster not in known_clusters:
            raise ValueError(
                f"Cluster {validated_cluster!r} is not in the known clusters list: {sorted(known_clusters)!r}",
            )

        return validated_cluster

    @validator("addresses")
    def _validate_addresses(cls, value, values):
        cluster = values.get("cluster")

        log.debug("|%s| Normalizing addresses %r names...", cls.__name__, value)

        validated_addresses = [cls.Slots.normalize_address(address, cluster) or address for address in value]
        if validated_addresses != value:
            log.debug("|%s| Got %r", cls.__name__, validated_addresses)

        cluster_addresses = set(cls.Slots.get_cluster_addresses(cluster) or [])
        unknown_addresses = set(validated_addresses) - cluster_addresses
        if cluster_addresses and unknown_addresses:
            raise ValueError(f"Cluster {cluster!r} does not contain addresses {unknown_addresses!r}")

        return validated_addresses

    @validator("spark")
    def _check_spark_version(cls, spark):
        spark_version = get_spark_version(spark)
        if spark_version < Version("2.4"):
            raise ValueError(f"Spark version must be at least 2.4, got {spark_version}")

        return spark

    @validator("spark")
    def _check_java_class_imported(cls, spark):
        java_class = "org.apache.spark.sql.kafka010.KafkaSourceProvider"

        try:
            try_import_java_class(spark, java_class)
        except Exception as e:
            spark_version = get_spark_version(spark).format("{major}.{minor}")
            msg = MISSING_JVM_CLASS_MSG.format(
                java_class=java_class,
                package_source=cls.__name__,
                args=f"spark_version='{spark_version}'",
            )
            if log.isEnabledFor(logging.DEBUG):
                log.debug("Missing Java class", exc_info=e, stack_info=True)
            raise ValueError(msg) from e
        return spark

    def _get_connection_properties(self) -> dict:
        result = {"bootstrap.servers": ",".join(self.addresses)}
        result.update(self.extra.dict(by_alias=True, exclude_none=True))
        result.update(self.protocol.get_options(self))
        if self.auth:
            result.update(self.auth.get_options(self))

        result["client.id"] = self.spark.sparkContext.appName  # type: ignore[assignment]
        return stringify(result)

    def _get_java_consumer(self):
        connection_properties = self._get_connection_properties()
        connection_properties.update(
            {
                # Consumer cannot be created without these options
                # They are set by Spark internally, but for manual consumer creation we need to pass them explicitly
                "key.deserializer": "org.apache.kafka.common.serialization.ByteArrayDeserializer",
                "value.deserializer": "org.apache.kafka.common.serialization.ByteArrayDeserializer",
            },
        )
        jvm = self.spark._jvm
        consumer_class = jvm.org.apache.kafka.clients.consumer.KafkaConsumer
        return consumer_class(connection_properties)

    def _get_topics(self, timeout: int = 10) -> set[str]:
        jvm = self.spark._jvm  # type: ignore[attr-defined]
        # Maybe we should not pass explicit timeout at all,
        # and instead use default.api.timeout.ms which is configurable via self.extra.
        # Think about this next time if someone see issues in real use
        duration = jvm.java.time.Duration.ofSeconds(timeout)  # type: ignore[union-attr]
        consumer = self._get_java_consumer()
        with closing(consumer):
            topics = consumer.listTopics(duration)
            return set(topics)

    def _log_parameters(self):
        log.info("|%s| Using connection parameters:", self.__class__.__name__)
        log_with_indent(log, "cluster = %r", self.cluster)
        log_collection(log, "addresses", self.addresses, max_items=10)
        log_with_indent(log, "protocol = %r", self.protocol)
        log_with_indent(log, "auth = %r", self.auth)
        log_with_indent(log, "extra = %r", self.extra.dict(by_alias=True, exclude_none=True))



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import logging
import os
import shutil
from typing import TYPE_CHECKING, Optional

try:
    from pydantic.v1 import Field, PrivateAttr, root_validator, validator
except (ImportError, AttributeError):
    from pydantic import Field, PrivateAttr, root_validator, validator  # type: ignore[no-redef, assignment]

from onetl._util.file import get_file_hash, is_file_readable
from onetl._util.spark import stringify
from onetl.connection.db_connection.kafka.kafka_auth import KafkaAuth
from onetl.impl import GenericOptions, LocalPath, path_repr

if TYPE_CHECKING:
    from onetl.connection import Kafka

log = logging.getLogger(__name__)


KNOWN_OPTIONS = frozenset(
    (
        "clearPass",
        "debug",
        "doNotPrompt",
        "isInitiator",
        "refreshKrb5Config",
        "renewTGT",
        "storePass",
        "ticketCache",
        "tryFirstPass",
        "useFirstPass",
        "sasl.kerberos.*",
    ),
)


PROHIBITED_OPTIONS = frozenset(
    (
        # Already set by class itself
        "sasl.kerberos.service.name",
        "sasl.jaas.config",
        "sasl.mechanism",
    ),
)


class KafkaKerberosAuth(KafkaAuth, GenericOptions):
    """
    Connect to Kafka using ``sasl.mechanism="GSSAPI"``.

    For more details see:

    * `Kafka Documentation <https://kafka.apache.org/documentation/#security_sasl_kerberos_clientconfig>`_
    * `Krb5LoginModule documentation <https://docs.oracle.com/javase/8/docs/jre/api/security/jaas/spec/com/sun/security/auth/module/Krb5LoginModule.html>`_

    .. versionadded:: 0.9.0

    Examples
    --------

    Auth in Kafka with keytab, automatically deploy keytab files to all Spark hosts (driver and executors):

    .. code:: python

        from onetl.connection import Kafka

        auth = Kafka.KerberosAuth(
            principal="user",
            keytab="/path/to/keytab",
            deploy_keytab=True,
        )

    Auth in Kafka with keytab, keytab is **already deployed** on all Spark hosts (driver and executors):

    .. code:: python

        from onetl.connection import Kafka

        auth = Kafka.KerberosAuth(
            principal="user",
            keytab="/path/to/keytab",
            deploy_keytab=False,
        )

    Auth in Kafka with existing Kerberos ticket (only Spark session created with ``master=local``):

    .. code:: python

        from onetl.connection import Kafka

        auth = Kafka.KerberosAuth(
            principal="user",
            use_keytab=False,
            use_ticket_cache=True,
        )

    Pass custom options for JAAS config and Kafka SASL:

    .. code:: python

        from onetl.connection import Kafka

        auth = Kafka.KerberosAuth.parse(
            {
                "principal": "user",
                "keytab": "/path/to/keytab",
                # options without sasl.kerberos. prefix are passed to JAAS config
                # names are in camel case!
                "isInitiator": True,
                # options with `sasl.kerberos.` prefix are passed to Kafka client config as-is
                "sasl.kerberos.kinit.cmd": "/usr/bin/kinit",
            }
        )
    """

    principal: str
    keytab: Optional[LocalPath] = Field(default=None, alias="keyTab")
    deploy_keytab: bool = True
    service_name: str = Field(default="kafka", alias="serviceName")
    renew_ticket: bool = Field(default=True, alias="renewTicket")
    store_key: bool = Field(default=True, alias="storeKey")
    use_keytab: bool = Field(default=True, alias="useKeyTab")
    use_ticket_cache: bool = Field(default=False, alias="useTicketCache")

    _keytab_path: Optional[LocalPath] = PrivateAttr(default=None)

    class Config:
        prohibited_options = PROHIBITED_OPTIONS
        known_options = KNOWN_OPTIONS
        strip_prefixes = ["kafka."]
        extra = "allow"

    def get_jaas_conf(self, kafka: Kafka) -> str:
        options = self.dict(
            by_alias=True,
            exclude_none=True,
            exclude={"deploy_keytab"},
        )
        if self.keytab:
            options["keyTab"] = self._prepare_keytab(kafka)

        jaas_conf = stringify({key: value for key, value in options.items() if not key.startswith("sasl.")}, quote=True)
        jaas_conf_items = [f"{key}={value}" for key, value in jaas_conf.items()]
        return "com.sun.security.auth.module.Krb5LoginModule required " + " ".join(jaas_conf_items) + ";"

    def get_options(self, kafka: Kafka) -> dict:
        result = {
            key: value for key, value in self.dict(by_alias=True, exclude_none=True).items() if key.startswith("sasl.")
        }
        result.update(
            {
                "sasl.mechanism": "GSSAPI",
                "sasl.jaas.config": self.get_jaas_conf(kafka),
                "sasl.kerberos.service.name": self.service_name,
            },
        )
        return stringify(result)

    def cleanup(self, kafka: Kafka) -> None:
        if self._keytab_path and self._keytab_path.exists():
            log.debug("Removing keytab from %s", path_repr(self._keytab_path))
            try:
                self._keytab_path.unlink()
            except OSError:
                log.exception("Failed to remove keytab file '%s'", self._keytab_path)
        self._keytab_path = None

    @validator("keytab")
    def _validate_keytab(cls, value):
        return is_file_readable(value)

    @root_validator
    def _use_keytab(cls, values):
        keytab = values.get("keytab")
        use_keytab = values.get("use_keytab")
        if use_keytab and not keytab:
            raise ValueError("keytab is required if useKeytab is True")
        return values

    def _prepare_keytab(self, kafka: Kafka) -> str:
        keytab: LocalPath = self.keytab  # type: ignore[assignment]
        if not self.deploy_keytab:
            return os.fspath(keytab)

        self._keytab_path = self._generate_keytab_path(keytab, self.principal)
        log.debug("Moving keytab from %s to %s", path_repr(keytab), path_repr(self._keytab_path))
        shutil.copy2(keytab, self._keytab_path)
        kafka.spark.sparkContext.addFile(os.fspath(self._keytab_path))
        return os.fspath(self._keytab_path.name)

    @staticmethod
    def _generate_keytab_path(keytab: LocalPath, principal: str) -> LocalPath:
        # Using hash in keytab name prevents collisions if there are several Kafka instances with
        # keytab paths like `/some/kafka.keytab` and `/another/kafka.keytab`.
        # Not using random values here to avoid generating too much garbage in current directory.
        # Also Spark copies all files to the executor working directory ignoring their paths,
        # so if there is some other connector deploy keytab file, these files will have different names.
        keytab_hash = get_file_hash(keytab, "md5")
        return LocalPath().joinpath(f"kafka_{principal}_{keytab_hash.hexdigest()}.keytab").resolve()



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from onetl.connection.db_connection.mysql.connection import MySQL, MySQLExtra
from onetl.connection.db_connection.mysql.dialect import MySQLDialect



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from onetl.impl import GenericOptions

PROHIBITED_OPTIONS = frozenset(
    (
        # filled by onETL classes
        "bootstrap.servers",
        "security.protocol",
        "sasl.*",
        "ssl.*",
        # Not supported by Spark
        "auto.offset.reset",
        "enable.auto.commit",
        "interceptor.classes",
        "key.deserializer",
        "key.serializer",
        "value.deserializer",
        "value.serializer",
    ),
)


class KafkaExtra(GenericOptions):
    """
    This class is responsible for validating additional options that are passed from the user
    to the Kafka connection. These extra options are configurations that can be provided to the
    Kafka, which aren't part of the core connection options.

    See Connection `producer options documentation <https://kafka.apache.org/documentation/#producerconfigs>`_,
    `consumer options documentation <https://kafka.apache.org/documentation/#consumerconfigs>`_
    for more details
    """

    class Config:
        strip_prefixes = ["kafka."]
        prohibited_options = PROHIBITED_OPTIONS
        extra = "allow"



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from onetl.connection.file_connection.hdfs.connection import HDFS
from onetl.connection.file_connection.hdfs.slots import HDFSSlots



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from onetl.file.file_uploader.file_uploader import FileUploader
from onetl.file.file_uploader.options import FileUploaderOptions
from onetl.file.file_uploader.result import UploadResult



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from onetl.connection.db_connection.teradata.connection import Teradata, TeradataExtra
from onetl.connection.db_connection.teradata.dialect import TeradataDialect



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, ClassVar

try:
    from pydantic.v1 import Field
except (ImportError, AttributeError):
    from pydantic import Field  # type: ignore[no-redef, assignment]

from onetl._util.spark import get_spark_version, stringify
from onetl.file.format.file_format import ReadWriteFileFormat
from onetl.hooks import slot, support_hooks

if TYPE_CHECKING:
    from pyspark.sql import Column, SparkSession
    from pyspark.sql.types import StructType


READ_WRITE_OPTIONS = frozenset(
    (
        "charToEscapeQuoteEscaping",
        "dateFormat",
        "emptyValue",
        "ignoreLeadingWhiteSpace",
        "ignoreTrailingWhiteSpace",
        "nullValue",
        "timestampFormat",
        "timestampNTZFormat",
    ),
)

READ_OPTIONS = frozenset(
    (
        "columnNameOfCorruptRecord",
        "comment",
        "enableDateTimeParsingFallback",
        "enforceSchema",
        "inferSchema",
        "locale",
        "maxCharsPerColumn",
        "maxColumns",
        "mode",
        "multiLine",
        "nanValue",
        "negativeInf",
        "positiveInf",
        "preferDate",
        "samplingRatio",
        "unescapedQuoteHandling",
    ),
)

WRITE_OPTIONS = frozenset(
    (
        "compression",
        "escapeQuotes",
        "quoteAll",
    ),
)


@support_hooks
class CSV(ReadWriteFileFormat):
    """
    CSV file format. |support_hooks|

    Based on `Spark CSV <https://spark.apache.org/docs/latest/sql-data-sources-csv.html>`_ file format.

    Supports reading/writing files with ``.csv`` extension with content like:

    .. code-block:: csv
        :caption: example.csv

        "some","value"
        "another","value"

    .. note ::

        You can pass any option to the constructor, even if it is not mentioned in this documentation.
        **Option names should be in** ``camelCase``!

        The set of supported options depends on Spark version. See link above.

    .. versionadded:: 0.9.0

    Examples
    --------

    Describe options how to read from/write to CSV file with specific options:

    .. code:: python

        csv = CSV(sep=",", encoding="utf-8", inferSchema=True, compression="gzip")

    """

    name: ClassVar[str] = "csv"
    delimiter: str = Field(default=",", alias="sep")
    encoding: str = "utf-8"
    quote: str = '"'
    escape: str = "\\"
    header: bool = False
    lineSep: str = "\n"  # noqa: N815

    class Config:
        known_options = READ_WRITE_OPTIONS | READ_OPTIONS | WRITE_OPTIONS
        extra = "allow"

    @slot
    @classmethod
    def check_if_supported(cls, spark: SparkSession) -> None:
        # always available
        pass

    def parse_column(self, column: str | Column, schema: StructType) -> Column:
        """
        Parses a CSV string column to a structured Spark SQL column using Spark's
        `from_csv <https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.functions.from_csv.html>`_ function, based on the provided schema.

        .. note::

            Can be used only with Spark 3.x+

        .. versionadded:: 0.11.0

        Parameters
        ----------
        column : str | Column
            The name of the column or the column object containing CSV strings/bytes to parse.

        schema : StructType
            The schema to apply when parsing the CSV data. This defines the structure of the output DataFrame column.

        Returns
        -------
        Column with deserialized data, with the same structure as the provided schema. Column name is the same as input column.

        Examples
        --------

        >>> from pyspark.sql.types import StructType, StructField, IntegerType, StringType
        >>> from onetl.file.format import CSV
        >>> df.show()
        +--+--------+
        |id|value   |
        +--+--------+
        |1 |Alice;20|
        |2 |Bob;25  |
        +--+--------+
        >>> df.printSchema()
        root
        |-- id: integer (nullable = true)
        |-- value: string (nullable = true)
        >>> csv = CSV(delimiter=";")
        >>> csv_schema = StructType(
        ...     [
        ...         StructField("name", StringType(), nullable=True),
        ...         StructField("age", IntegerType(), nullable=True),
        ...     ],
        ... )
        >>> parsed_df = df.select("id", csv.parse_column("value", csv_schema))
        >>> parsed_df.show()
        +--+-----------+
        |id|value      |
        +--+-----------+
        |1 |{Alice, 20}|
        |2 |  {Bob, 25}|
        +--+-----------+
        >>> parsed_df.printSchema()
        root
        |-- id: integer (nullable = true)
        |-- value: struct (nullable = true)
        |    |-- name: string (nullable = true)
        |    |-- age: integer (nullable = true)
        """

        from pyspark.sql import Column, SparkSession  # noqa: WPS442

        spark = SparkSession._instantiatedSession  # noqa: WPS437
        self.check_if_supported(spark)
        self._check_spark_version_for_serialization(spark)
        self._check_unsupported_serialization_options()

        from pyspark.sql.functions import col, from_csv

        if isinstance(column, Column):
            column_name = column._jc.toString()  # noqa: WPS437
        else:
            column_name, column = column, col(column).cast("string")

        schema_string = schema.simpleString()
        options = stringify(self.dict(by_alias=True))
        return from_csv(column, schema_string, options).alias(column_name)

    def serialize_column(self, column: str | Column) -> Column:
        """
        Serializes a structured Spark SQL column into a CSV string column using Spark's
        `to_csv <https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.functions.to_csv.html>`_ function.

        .. note::

            Can be used only with Spark 3.x+

        .. versionadded:: 0.11.0

        Parameters
        ----------
        column : str | Column
            The name of the column or the Column object containing the data to serialize to CSV.

        Returns
        -------
        Column with string CSV data. Column name is the same as input column.

        Examples
        --------

        >>> from pyspark.sql.functions import decode
        >>> from onetl.file.format import CSV
        >>> df.show()
        +--+-----------+
        |id|value      |
        +--+-----------+
        |1 |{Alice, 20}|
        |2 |  {Bob, 25}|
        +--+-----------+
        >>> df.printSchema()
        root
        |-- id: integer (nullable = true)
        |-- value: struct (nullable = true)
        |    |-- name: string (nullable = true)
        |    |-- age: integer (nullable = true)
        >>> # serializing data into CSV format
        >>> csv = CSV(delimiter=";")
        >>> serialized_df = df.select("id", csv.serialize_column("value"))
        >>> serialized_df.show(truncate=False)
        +--+--------+
        |id|value   |
        +--+--------+
        |1 |Alice;20|
        |2 |Bob;25  |
        +--+--------+
        >>> serialized_df.printSchema()
        root
        |-- id: integer (nullable = true)
        |-- value: string (nullable = true)
        """

        from pyspark.sql import Column, SparkSession  # noqa: WPS442

        spark = SparkSession._instantiatedSession  # noqa: WPS437
        self.check_if_supported(spark)
        self._check_spark_version_for_serialization(spark)
        self._check_unsupported_serialization_options()

        from pyspark.sql.functions import col, to_csv

        if isinstance(column, Column):
            column_name = column._jc.toString()  # noqa: WPS437
        else:
            column_name, column = column, col(column)

        options = stringify(self.dict(by_alias=True))
        return to_csv(column, options).alias(column_name)

    def _check_spark_version_for_serialization(self, spark: SparkSession):
        spark_version = get_spark_version(spark)
        if spark_version.major < 3:
            class_name = self.__class__.__name__
            error_msg = (
                f"`{class_name}.parse_column` or `{class_name}.serialize_column` are available "
                f"only since Spark 3.x, but got {spark_version}"
            )
            raise ValueError(error_msg)

    def _check_unsupported_serialization_options(self):
        unsupported_options = ["header", "compression", "inferSchema"]
        current_options = self.dict()
        for option in unsupported_options:
            if current_options.get(option):
                warnings.warn(
                    f"Option `{option}` is set but not supported in `CSV.parse_column` or `CSV.serialize_column`. "
                    "This may lead to unexpected behavior.",
                    UserWarning,
                    stacklevel=2,
                )



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum

from onetl._metrics.listener.stage import SparkListenerStage, SparkListenerStageStatus
from onetl._util.scala import scala_seq_to_python_list


class SparkListenerJobStatus(str, Enum):
    RUNNING = "RUNNING"
    SUCCEEDED = "SUCCEEDED"
    FAILED = "FAILED"
    UNKNOWN = "UNKNOWN"

    def __str__(self):
        return self.value


@dataclass
class SparkListenerJob:
    id: int
    description: str | None = None
    group_id: str | None = None
    call_site: str | None = None
    status: SparkListenerJobStatus = SparkListenerJobStatus.UNKNOWN

    _stages: dict[int, SparkListenerStage] = field(default_factory=dict, repr=False, init=False)

    @property
    def stages(self) -> list[SparkListenerStage]:
        result = []
        for stage_id in sorted(self._stages.keys()):
            result.append(self._stages[stage_id])
        return result

    @classmethod
    def create(cls, event):
        # https://spark.apache.org/docs/3.5.2/api/java/org/apache/spark/scheduler/SparkListenerJobSubmitted.html
        # https://spark.apache.org/docs/3.5.2/api/java/org/apache/spark/scheduler/SparkListenerJobCompleted.html
        result = cls(
            id=event.jobId(),
            description=event.properties().get("spark.job.description"),
            group_id=event.properties().get("spark.jobGroup.id"),
            call_site=event.properties().get("callSite.short"),
        )

        stage_ids = scala_seq_to_python_list(event.stageIds())
        stage_infos = scala_seq_to_python_list(event.stageInfos())
        for stage_id, stage_info in zip(stage_ids, stage_infos):
            result._stages[stage_id] = SparkListenerStage.create(stage_info)  # noqa: WPS437

        return result

    def on_job_start(self, event):
        self.status = SparkListenerJobStatus.RUNNING

    def on_job_end(self, event):
        for stage in self._stages.values():
            if stage.status == SparkListenerStageStatus.FAILED:
                self.status = SparkListenerJobStatus.FAILED
                break
        else:
            self.status = SparkListenerJobStatus.SUCCEEDED

    def on_stage_start(self, event):
        stage_id = event.stageInfo().stageId()
        stage = self._stages.get(stage_id)
        if stage:
            stage.on_stage_start(event)

    def on_stage_end(self, event):
        stage_id = event.stageInfo().stageId()
        stage = self._stages.get(stage_id)
        if stage:
            stage.on_stage_end(event)

    # push down events
    def on_task_start(self, event):
        for stage in self._stages.values():
            stage.on_task_start(event)

    def on_task_end(self, event):
        for stage in self._stages.values():
            stage.on_task_end(event)



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from onetl.file.file_mover.file_mover import FileMover
from onetl.file.file_mover.options import FileMoverOptions
from onetl.file.file_mover.result import MoveResult



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations


class SupportNameAny:
    def validate_name(self, value: str) -> str:
        return value



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from onetl.hooks.hook import HookPriority, hook
from onetl.hooks.hooks_state import resume_all_hooks, skip_all_hooks, stop_all_hooks
from onetl.hooks.slot import slot
from onetl.hooks.support_hooks import support_hooks



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from onetl.file.file_df_reader import FileDFReader
from onetl.file.file_df_writer import FileDFWriter
from onetl.file.file_downloader import DownloadResult, FileDownloader
from onetl.file.file_mover import FileMover, MoveResult
from onetl.file.file_uploader import FileUploader, UploadResult
